from pylabrobot.resources import Coordinate
from pylabrobot.resources.carrier import PlateCarrier, PlateHolder


def _liconic_rack(
  name: str,
  site_height: int,
  num_sites: int,
  model: str,
  total_height: int = 505,
):
  start = 17.2  # rough height of first plate position
  return PlateCarrier(
    name=name,
    size_x=109,  # based off cytomat rack dimensions roughly the same
    size_y=142,
    size_z=total_height,
    sites={
      i: PlateHolder(
        size_x=85.48,
        size_y=127.27,
        # estimates
        size_z=max(site_height, total_height - site_height) if i == num_sites - 1 else site_height,
        name=f"{name}-{i}",
        pedestal_size_z=0,
      ).at(
        Coordinate(
          x=11.76,  # estimate
          y=0,
          z=start + site_height * i,
        )
      )
      for i in range(num_sites)
    },
    model=model,
  )


def liconic_rack_5mm_42(name: str):
  """STX44. Pitch 11mm, motor steps 377."""
  return _liconic_rack(name=name, site_height=5, num_sites=42, model="liconic_rack_5mm_42")


def liconic_rack_5mm_55(name: str):
  """STX500 bicarousel. Pitch 11mm, motor steps 377."""
  return _liconic_rack(
    name=name,
    site_height=5,
    num_sites=55,
    model="liconic_rack_5mm_55",
    total_height=645,
  )


def liconic_rack_5mm_111(name: str):
  """STX1000 bicarousel. Pitch 11mm, motor steps 377."""
  return _liconic_rack(
    name=name,
    site_height=5,
    num_sites=111,
    model="liconic_rack_5mm_111",
    total_height=1210,
  )


def liconic_rack_11mm_28(name: str):
  """STX44. Pitch 17mm, motor steps 582."""
  return _liconic_rack(name=name, site_height=11, num_sites=28, model="liconic_rack_11mm_28")


def liconic_rack_11mm_37(name: str):
  """STX500 bicarousel. Pitch 17mm, motor steps 582."""
  return _liconic_rack(
    name=name,
    site_height=11,
    num_sites=37,
    model="liconic_rack_11mm_37",
    total_height=645,
  )


def liconic_rack_11mm_72(name: str):
  """STX1000 bicarousel. Pitch 17mm, motor steps 582."""
  return _liconic_rack(
    name=name,
    site_height=11,
    num_sites=72,
    model="liconic_rack_11mm_72",
    total_height=1210,
  )


def liconic_rack_12mm_27(name: str):
  """STX44. Pitch 18mm, motor steps 617."""
  return _liconic_rack(name=name, site_height=12, num_sites=27, model="liconic_rack_12mm_27")


def liconic_rack_12mm_35(name: str):
  """STX500 bicarousel. Pitch 18mm, motor steps 617."""
  return _liconic_rack(
    name=name,
    site_height=12,
    num_sites=35,
    model="liconic_rack_12mm_35",
    total_height=645,
  )


def liconic_rack_12mm_68(name: str):
  """STX1000 bicarousel. Pitch 18mm, motor steps 617."""
  return _liconic_rack(
    name=name,
    site_height=12,
    num_sites=68,
    model="liconic_rack_12mm_68",
    total_height=1210,
  )


def liconic_rack_17mm_22(name: str):
  """STX44. Pitch 23mm, motor steps 788."""
  return _liconic_rack(name=name, site_height=17, num_sites=22, model="liconic_rack_17mm_22")


def liconic_rack_17mm_28(name: str):
  """STX500 bicarousel. Pitch 23mm, motor steps 788."""
  return _liconic_rack(
    name=name,
    site_height=17,
    num_sites=28,
    model="liconic_rack_17mm_28",
    total_height=645,
  )


def liconic_rack_17mm_53(name: str):
  """STX1000 bicarousel. Pitch 23mm, motor steps 788."""
  return _liconic_rack(
    name=name,
    site_height=17,
    num_sites=53,
    model="liconic_rack_17mm_53",
    total_height=1210,
  )


def liconic_rack_22mm_17(name: str):
  """STX44. Pitch 28mm, motor steps 959."""
  return _liconic_rack(name=name, site_height=22, num_sites=17, model="liconic_rack_22mm_17")


def liconic_rack_22mm_23(name: str):
  """STX500 bicarousel. Pitch 28mm, motor steps 959."""
  return _liconic_rack(
    name=name,
    site_height=22,
    num_sites=23,
    model="liconic_rack_22mm_23",
    total_height=645,
  )


def liconic_rack_22mm_43(name: str):
  """STX1000 bicarousel. Pitch 28mm, motor steps 959."""
  return _liconic_rack(
    name=name,
    site_height=22,
    num_sites=43,
    model="liconic_rack_22mm_43",
    total_height=1210,
  )


def liconic_rack_23mm_17(name: str):
  """STX44. Pitch 29mm, motor steps 994."""
  return _liconic_rack(name=name, site_height=23, num_sites=17, model="liconic_rack_23mm_17")


def liconic_rack_23mm_22(name: str):
  """STX500 bicarousel. Pitch 29mm, motor steps 994."""
  return _liconic_rack(
    name=name,
    site_height=23,
    num_sites=22,
    model="liconic_rack_23mm_22",
    total_height=645,
  )


def liconic_rack_23mm_42(name: str):
  """STX1000 bicarousel. Pitch 29mm, motor steps 994."""
  return _liconic_rack(
    name=name,
    site_height=23,
    num_sites=42,
    model="liconic_rack_23mm_42",
    total_height=1210,
  )


def liconic_rack_24mm_17(name: str):
  """STX44. Pitch 30mm, motor steps 1028."""
  return _liconic_rack(name=name, site_height=24, num_sites=17, model="liconic_rack_24mm_17")


def liconic_rack_24mm_21(name: str):
  """STX500 bicarousel. Pitch 30mm, motor steps 1028."""
  return _liconic_rack(
    name=name,
    site_height=24,
    num_sites=21,
    model="liconic_rack_24mm_21",
    total_height=645,
  )


def liconic_rack_24mm_41(name: str):
  """STX1000 bicarousel. Pitch 30mm, motor steps 1028."""
  return _liconic_rack(
    name=name,
    site_height=24,
    num_sites=41,
    model="liconic_rack_24mm_41",
    total_height=1210,
  )


def liconic_rack_27mm_15(name: str):
  """STX44. Pitch 33mm, motor steps 1131."""
  return _liconic_rack(name=name, site_height=27, num_sites=15, model="liconic_rack_27mm_15")


def liconic_rack_27mm_19(name: str):
  """STX500 bicarousel. Pitch 33mm, motor steps 1131."""
  return _liconic_rack(
    name=name,
    site_height=27,
    num_sites=19,
    model="liconic_rack_27mm_19",
    total_height=645,
  )


def liconic_rack_27mm_37(name: str):
  """STX1000 bicarousel. Pitch 33mm, motor steps 1131."""
  return _liconic_rack(
    name=name,
    site_height=27,
    num_sites=37,
    model="liconic_rack_27mm_37",
    total_height=1210,
  )


def liconic_rack_44mm_10(name: str):
  """STX44. Pitch 50mm, motor steps 1713."""
  return _liconic_rack(name=name, site_height=44, num_sites=10, model="liconic_rack_44mm_10")


def liconic_rack_44mm_13(name: str):
  """STX500 bicarousel. Pitch 50mm, motor steps 1713."""
  return _liconic_rack(
    name=name,
    site_height=44,
    num_sites=13,
    model="liconic_rack_44mm_13",
    total_height=645,
  )


def liconic_rack_44mm_25(name: str):
  """STX1000 bicarousel. Pitch 50mm, motor steps 1713."""
  return _liconic_rack(
    name=name,
    site_height=44,
    num_sites=25,
    model="liconic_rack_44mm_25",
    total_height=1210,
  )


def liconic_rack_53mm_8(name: str):
  """STX44. Pitch 59mm, motor steps 2021."""
  return _liconic_rack(name=name, site_height=53, num_sites=8, model="liconic_rack_53mm_8")


def liconic_rack_53mm_10(name: str):
  """STX500 bicarousel. Pitch 59mm, motor steps 2021."""
  return _liconic_rack(
    name=name,
    site_height=53,
    num_sites=10,
    model="liconic_rack_53mm_10",
    total_height=645,
  )


def liconic_rack_53mm_21(name: str):
  """STX1000 bicarousel. Pitch 59mm, motor steps 2021."""
  return _liconic_rack(
    name=name,
    site_height=53,
    num_sites=21,
    model="liconic_rack_53mm_21",
    total_height=1210,
  )


def liconic_rack_66mm_7(name: str):
  """STX44. Pitch 72mm, motor steps 2467."""
  return _liconic_rack(name=name, site_height=66, num_sites=7, model="liconic_rack_66mm_7")


def liconic_rack_66mm_8(name: str):
  """STX500 bicarousel. Pitch 72mm, motor steps 2467."""
  return _liconic_rack(
    name=name,
    site_height=66,
    num_sites=8,
    model="liconic_rack_66mm_8",
    total_height=645,
  )


def liconic_rack_66mm_17(name: str):
  """STX1000 bicarousel. Pitch 72mm, motor steps 2467."""
  return _liconic_rack(
    name=name,
    site_height=66,
    num_sites=17,
    model="liconic_rack_66mm_17",
    total_height=1210,
  )


def liconic_rack_104mm_4(name: str):
  """STX44. Pitch 110mm, motor steps 3563."""
  return _liconic_rack(name=name, site_height=104, num_sites=4, model="liconic_rack_104mm_4")


def liconic_rack_104mm_5(name: str):
  """STX500 bicarousel. Pitch 110mm, motor steps 3563."""
  return _liconic_rack(
    name=name,
    site_height=104,
    num_sites=5,
    model="liconic_rack_104mm_5",
    total_height=645,
  )


def liconic_rack_104mm_11(name: str):
  """STX1000 bicarousel. Pitch 110mm, motor steps 3563."""
  return _liconic_rack(
    name=name,
    site_height=104,
    num_sites=11,
    model="liconic_rack_104mm_11",
    total_height=1210,
  )



import abc
import time
import warnings
from enum import Enum
from typing import Dict, Literal, Optional

from pylabrobot.heating_shaking.backend import HeaterShakerBackend
from pylabrobot.io.usb import USB


class PlateLockPosition(Enum):
  LOCKED = 1
  UNLOCKED = 0


class HamiltonHeaterShakerInterface(abc.ABC):
  """Either a control box or a STAR: the api is the same"""

  @abc.abstractmethod
  async def send_hhs_command(self, index: int, command: str, **kwargs) -> str:
    pass


class HamiltonHeaterShakerBox(HamiltonHeaterShakerInterface):
  def __init__(
    self,
    id_vendor: int = 0x8AF,
    id_product: int = 0x8002,
    device_address: Optional[int] = None,
    serial_number: Optional[str] = None,
  ):
    self.io = USB(
      human_readable_device_name="Hamilton Heater Shaker Box",
      id_vendor=id_vendor,
      id_product=id_product,
      device_address=device_address,
      serial_number=serial_number,
    )
    self._id = 0

  def _generate_id(self) -> int:
    """continuously generate unique ids 0 <= x < 10000."""
    self._id += 1
    return self._id % 10000

  async def setup(self):
    """
    If io.setup() fails, ensure that libusb drivers were installed for the HHS as per docs.
    """
    await self.io.setup()

  async def stop(self):
    await self.io.stop()

  async def send_hhs_command(self, index: int, command: str, **kwargs) -> str:
    args = "".join([f"{key}{value}" for key, value in kwargs.items()])
    id_ = str(self._generate_id()).zfill(4)
    await self.io.write(f"T{index}{command}id{id_}{args}".encode())
    return (await self.io.read()).decode("utf-8")


class HamiltonHeaterShakerBackend(HeaterShakerBackend):
  """Backend for Hamilton Heater Shaker devices connected through an Heater Shaker Box"""

  @property
  def supports_active_cooling(self) -> bool:
    return False

  def __init__(self, index: int, interface: HamiltonHeaterShakerInterface) -> None:
    """
    Multiple Hamilton Heater Shakers can be connected to the same Heat Shaker Box. Each has A
    unique 'shaker index'
    """
    assert index >= 0, "Shaker index must be non-negative"
    self.index = index

    super().__init__()
    self.interface = interface

  async def setup(self):
    """
    If io.setup() fails, ensure that libusb drivers were installed for the HHS as per docs.
    """
    await self._initialize_lock()
    await self._initialize_shaker_drive()

  async def stop(self):
    pass

  def serialize(self) -> dict:
    warnings.warn("The interface is not serialized.")

    heater_shaker_serialized = HeaterShakerBackend.serialize(self)
    return {
      **heater_shaker_serialized,
      "index": self.index,
      "interface": None,  # TODO: implement serialization
    }

  async def start_shaking(
    self,
    speed: float = 800,
    direction: Literal[0, 1] = 0,
    acceleration: int = 1_000,
    timeout: Optional[float] = 30,
  ):
    """
    if the plate is not locked, it will be locked.

    speed: steps per second
    direction: 0 for positive, 1 for negative
    acceleration: increments per second
    """

    await self.lock_plate()

    int_speed = int(speed)
    assert 20 <= int_speed <= 2_000, "Speed must be between 20 and 2_000"
    assert direction in [0, 1], "Direction must be 0 or 1"
    assert 500 <= acceleration <= 10_000, "Acceleration must be between 500 and 10_000"

    now = time.time()
    while True:
      await self._start_shaking(direction=direction, speed=int_speed, acceleration=acceleration)
      if await self.get_is_shaking():
        break
      if timeout is not None and time.time() - now > timeout:
        raise TimeoutError("Failed to start shaking within timeout")

  async def shake(
    self,
    speed: float = 800,
    direction: Literal[0, 1] = 0,
    acceleration: int = 1_000,
    timeout: Optional[float] = 30,
  ):
    warnings.warn(
      "HamiltonHeaterShakerBackend.shake() is deprecated and will be removed in a future release. "
      "Use start_shaking() instead.",
      DeprecationWarning,
      stacklevel=2,
    )
    await self.start_shaking(
      speed=speed,
      direction=direction,
      acceleration=acceleration,
      timeout=timeout,
    )

  async def stop_shaking(self):
    await self._stop_shaking()
    await self._wait_for_stop()

  async def get_is_shaking(self) -> bool:
    response = await self.interface.send_hhs_command(index=self.index, command="RD")
    return response.endswith("1")  # type: ignore[no-any-return] # what

  async def _move_plate_lock(self, position: PlateLockPosition):
    return await self.interface.send_hhs_command(index=self.index, command="LP", lp=position.value)

  @property
  def supports_locking(self) -> bool:
    return True

  async def lock_plate(self):
    await self._move_plate_lock(PlateLockPosition.LOCKED)

  async def unlock_plate(self):
    await self._move_plate_lock(PlateLockPosition.UNLOCKED)

  async def _initialize_lock(self):
    """Firmware command initialize lock."""
    return await self.interface.send_hhs_command(index=self.index, command="LI")

  async def _initialize_shaker_drive(self):
    """Initialize the shaker drive, homing to absolute position 0"""
    return await self.interface.send_hhs_command(index=self.index, command="SI")

  async def _start_shaking(self, direction: int, speed: int, acceleration: int):
    """Firmware command for starting shaking."""
    speed_str = str(speed).zfill(4)
    acceleration_str = str(acceleration).zfill(5)
    return await self.interface.send_hhs_command(
      index=self.index, command="SB", st=direction, sv=speed_str, sr=acceleration_str
    )

  async def _stop_shaking(self):
    """Firmware command for stopping shaking."""
    return await self.interface.send_hhs_command(index=self.index, command="SC")

  async def _wait_for_stop(self):
    """Firmware command for waiting for shaking to stop."""
    return await self.interface.send_hhs_command(index=self.index, command="SW")

  async def set_temperature(self, temperature: float):
    """set temperature in Celsius"""
    assert 0 < temperature <= 105
    temp_str = f"{round(10 * temperature):04d}"
    return await self.interface.send_hhs_command(index=self.index, command="TA", ta=temp_str)

  async def _get_current_temperature(self) -> Dict[str, float]:
    """get temperature in Celsius"""
    response = await self.interface.send_hhs_command(index=self.index, command="RT")
    response = response.split("rt")[1]
    middle_temp = float(str(response).split(" ")[0].strip("+")) / 10
    edge_temp = float(str(response).split(" ")[1].strip("+")) / 10
    return {"middle": middle_temp, "edge": edge_temp}

  async def get_current_temperature(self) -> float:
    """get temperature in Celsius"""
    response = await self._get_current_temperature()
    return response["middle"]

  async def get_edge_temperature(self) -> float:
    """get temperature in Celsius"""
    response = await self._get_current_temperature()
    return response["edge"]

  async def deactivate(self):
    """turn off heating"""
    return await self.interface.send_hhs_command(index=self.index, command="TO")



import math
import unittest
from unittest.mock import AsyncMock, MagicMock, call, patch

from pylabrobot.plate_reading.molecular_devices.backend import (
  Calibrate,
  CarriageSpeed,
  KineticSettings,
  MolecularDevicesBackend,
  MolecularDevicesError,
  MolecularDevicesSettings,
  MolecularDevicesUnrecognizedCommandError,
  PmtGain,
  ReadMode,
  ReadOrder,
  ReadType,
  ShakeSettings,
  SpectrumSettings,
)
from pylabrobot.resources.agenbio.plates import AGenBio_96_wellplate_Ub_2200ul


class TestMolecularDevicesBackend(unittest.IsolatedAsyncioTestCase):
  backend: MolecularDevicesBackend
  mock_serial: MagicMock
  send_command_mock: AsyncMock

  def setUp(self):
    self.mock_serial = MagicMock()
    self.mock_serial.setup = AsyncMock()
    self.mock_serial.stop = AsyncMock()
    self.mock_serial.write = AsyncMock()
    self.mock_serial.readline = AsyncMock(return_value=b"OK>\r\n")

    with patch("pylabrobot.io.serial.Serial", return_value=self.mock_serial):
      self.backend = MolecularDevicesBackend(port="COM1")
      self.backend.io = self.mock_serial
      self.send_command_mock = patch.object(
        self.backend, "send_command", new_callable=AsyncMock
      ).start()
    self.addCleanup(patch.stopall)

  async def test_setup_stop(self):
    # un-mock send_command for this test
    with patch.object(
      self.backend, "send_command", wraps=self.backend.send_command
    ) as wrapped_send_command:
      await self.backend.setup()
      self.mock_serial.setup.assert_called_once()
      wrapped_send_command.assert_called_with("!")
      await self.backend.stop()
      self.mock_serial.stop.assert_called_once()

  async def test_set_clear(self):
    await self.backend._set_clear()
    self.send_command_mock.assert_called_once_with("!CLEAR DATA")

  async def test_set_mode(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_mode(settings)
    self.send_command_mock.assert_called_once_with("!MODE ENDPOINT")

    self.send_command_mock.reset_mock()
    settings.read_type = ReadType.KINETIC
    settings.kinetic_settings = KineticSettings(interval=10, num_readings=5)
    await self.backend._set_mode(settings)
    self.send_command_mock.assert_called_once_with("!MODE KINETIC 10 5")

    self.send_command_mock.reset_mock()
    settings.read_type = ReadType.SPECTRUM
    settings.spectrum_settings = SpectrumSettings(start_wavelength=200, step=10, num_steps=50)
    await self.backend._set_mode(settings)
    self.send_command_mock.assert_called_once_with("!MODE SPECTRUM 200 10 50")

    self.send_command_mock.reset_mock()
    settings.spectrum_settings.excitation_emission_type = "EXSPECTRUM"
    await self.backend._set_mode(settings)
    self.send_command_mock.assert_called_once_with("!MODE EXSPECTRUM 200 10 50")

  async def test_set_wavelengths(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      wavelengths=[500, (600, True)],
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_wavelengths(settings)
    self.send_command_mock.assert_called_once_with("!WAVELENGTH 500 F600")

    self.send_command_mock.reset_mock()
    settings.path_check = True
    await self.backend._set_wavelengths(settings)
    self.send_command_mock.assert_called_once_with("!WAVELENGTH 500 F600 900 998")

    self.send_command_mock.reset_mock()
    settings.read_mode = ReadMode.FLU
    settings.excitation_wavelengths = [485]
    settings.emission_wavelengths = [520]
    await self.backend._set_wavelengths(settings)
    self.send_command_mock.assert_has_calls([call("!EXWAVELENGTH 485"), call("!EMWAVELENGTH 520")])

    self.send_command_mock.reset_mock()
    settings.read_mode = ReadMode.LUM
    settings.emission_wavelengths = [590]
    await self.backend._set_wavelengths(settings)
    self.send_command_mock.assert_called_once_with("!EMWAVELENGTH 590")

  async def test_set_plate_position(self):
    plate = AGenBio_96_wellplate_Ub_2200ul("test_plate")
    settings = MolecularDevicesSettings(
      plate=plate,
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_plate_position(settings)
    self.send_command_mock.assert_has_calls(
      [call("!XPOS 13.380 9.000 12"), call("!YPOS 12.240 9.000 8")]
    )

  async def test_set_strip(self):
    plate = AGenBio_96_wellplate_Ub_2200ul("test_plate")
    settings = MolecularDevicesSettings(
      plate=plate,
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_strip(settings)
    self.send_command_mock.assert_called_once_with("!STRIP 1 12")

  async def test_set_shake(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_shake(settings)
    self.send_command_mock.assert_called_once_with("!SHAKE OFF")

    self.send_command_mock.reset_mock()
    settings.shake_settings = ShakeSettings(before_read=True, before_read_duration=5)
    await self.backend._set_shake(settings)
    self.send_command_mock.assert_has_calls([call("!SHAKE ON"), call("!SHAKE 5 0 0 0 0")])

    self.send_command_mock.reset_mock()
    settings.shake_settings = ShakeSettings(between_reads=True, between_reads_duration=3)
    settings.kinetic_settings = KineticSettings(interval=10, num_readings=5)
    await self.backend._set_shake(settings)
    self.send_command_mock.assert_has_calls([call("!SHAKE ON"), call("!SHAKE 0 10 7 3 0")])

  async def test_set_carriage_speed(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_carriage_speed(settings)
    self.send_command_mock.assert_called_once_with("!CSPEED 8")
    self.send_command_mock.reset_mock()
    settings.carriage_speed = CarriageSpeed.SLOW
    await self.backend._set_carriage_speed(settings)
    self.send_command_mock.assert_called_once_with("!CSPEED 1")

  async def test_set_read_stage(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.FLU,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_read_stage(settings)
    self.send_command_mock.assert_called_once_with("!READSTAGE TOP")
    self.send_command_mock.reset_mock()
    settings.read_from_bottom = True
    await self.backend._set_read_stage(settings)
    self.send_command_mock.assert_called_once_with("!READSTAGE BOT")
    self.send_command_mock.reset_mock()
    settings.read_mode = ReadMode.ABS
    await self.backend._set_read_stage(settings)
    self.send_command_mock.assert_not_called()

  async def test_set_flashes_per_well(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.FLU,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      flashes_per_well=10,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_flashes_per_well(settings)
    self.send_command_mock.assert_called_once_with("!FPW 10")
    self.send_command_mock.reset_mock()
    settings.read_mode = ReadMode.ABS
    await self.backend._set_flashes_per_well(settings)
    self.send_command_mock.assert_not_called()

  async def test_set_pmt(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.FLU,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      pmt_gain=PmtGain.AUTO,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_pmt(settings)
    self.send_command_mock.assert_called_once_with("!AUTOPMT ON")
    self.send_command_mock.reset_mock()
    settings.pmt_gain = PmtGain.HIGH
    await self.backend._set_pmt(settings)
    self.send_command_mock.assert_has_calls([call("!AUTOPMT OFF"), call("!PMT HIGH")])
    self.send_command_mock.reset_mock()
    settings.pmt_gain = 9
    await self.backend._set_pmt(settings)
    self.send_command_mock.assert_has_calls([call("!AUTOPMT OFF"), call("!PMT 9")])
    self.send_command_mock.reset_mock()
    settings.read_mode = ReadMode.ABS
    await self.backend._set_pmt(settings)
    self.send_command_mock.assert_not_called()

  async def test_set_filter(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.FLU,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      cutoff_filters=[self.backend._get_cutoff_filter_index_from_wavelength(535), 9],
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_filter(settings)
    self.send_command_mock.assert_has_calls([call("!AUTOFILTER OFF"), call("!EMFILTER 8 9")])
    self.send_command_mock.reset_mock()
    settings.cutoff_filters = []
    await self.backend._set_filter(settings)
    self.send_command_mock.assert_called_once_with("!AUTOFILTER ON")
    self.send_command_mock.reset_mock()
    settings.read_mode = ReadMode.ABS
    settings.cutoff_filters = [515, 530]
    await self.backend._set_filter(settings)
    self.send_command_mock.assert_called_once_with("!AUTOFILTER ON")

  async def test_set_calibrate(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_calibrate(settings)
    self.send_command_mock.assert_called_once_with("!CALIBRATE ON")
    self.send_command_mock.reset_mock()
    settings.read_mode = ReadMode.FLU
    await self.backend._set_calibrate(settings)
    self.send_command_mock.assert_called_once_with("!PMTCAL ON")

  async def test_set_order(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_order(settings)
    self.send_command_mock.assert_called_once_with("!ORDER COLUMN")
    self.send_command_mock.reset_mock()
    settings.read_order = ReadOrder.WAVELENGTH
    await self.backend._set_order(settings)
    self.send_command_mock.assert_called_once_with("!ORDER WAVELENGTH")

  async def test_set_speed(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=True,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_speed(settings)
    self.send_command_mock.assert_called_once_with("!SPEED ON")
    self.send_command_mock.reset_mock()
    settings.speed_read = False
    await self.backend._set_speed(settings)
    self.send_command_mock.assert_called_once_with("!SPEED OFF")
    self.send_command_mock.reset_mock()
    settings.read_mode = ReadMode.FLU
    await self.backend._set_speed(settings)
    self.send_command_mock.assert_not_called()

  async def test_set_integration_time(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.TIME,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    await self.backend._set_integration_time(settings, 10, 100)
    self.send_command_mock.assert_has_calls([call("!COUNTTIMEDELAY 10"), call("!COUNTTIME 0.1")])
    self.send_command_mock.reset_mock()
    settings.read_mode = ReadMode.ABS
    await self.backend._set_integration_time(settings, 10, 100)
    self.send_command_mock.assert_not_called()

  async def test_set_nvram_polar(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.POLAR,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
      settling_time=5,
    )
    await self.backend._set_nvram(settings)
    self.send_command_mock.assert_called_once_with("!NVRAM FPSETTLETIME 5")

  async def test_set_nvram_other(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
      settling_time=10,
    )
    await self.backend._set_nvram(settings)
    self.send_command_mock.assert_called_once_with("!NVRAM CARCOL 100")
    self.send_command_mock.reset_mock()
    settings.settling_time = 110
    await self.backend._set_nvram(settings)
    self.send_command_mock.assert_called_once_with("!NVRAM CARCOL 110")

  async def test_set_tag(self):
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.POLAR,
      read_type=ReadType.KINETIC,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=KineticSettings(interval=10, num_readings=5),
      spectrum_settings=None,
    )
    await self.backend._set_tag(settings)
    self.send_command_mock.assert_called_once_with("!TAG ON")
    self.send_command_mock.reset_mock()
    settings.read_type = ReadType.ENDPOINT
    await self.backend._set_tag(settings)
    self.send_command_mock.assert_called_once_with("!TAG OFF")
    self.send_command_mock.reset_mock()
    settings.read_mode = ReadMode.ABS
    settings.read_type = ReadType.KINETIC
    await self.backend._set_tag(settings)
    self.send_command_mock.assert_called_once_with("!TAG OFF")

  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._wait_for_idle",
    new_callable=AsyncMock,
  )
  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._transfer_data",
    new_callable=AsyncMock,
    return_value="",
  )
  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._read_now",
    new_callable=AsyncMock,
  )
  async def test_read_absorbance(self, mock_read_now, mock_transfer_data, mock_wait_for_idle):
    plate = AGenBio_96_wellplate_Ub_2200ul("test_plate")
    await self.backend.read_absorbance(plate, [500])

    commands = [c.args[0] for c in self.send_command_mock.call_args_list]
    self.assertIn("!CLEAR DATA", commands)
    self.assertIn("!STRIP 1 12", commands)
    self.assertIn("!CSPEED 8", commands)
    self.assertIn("!SHAKE OFF", commands)
    self.assertIn("!WAVELENGTH 500", commands)
    self.assertIn("!CALIBRATE ONCE", commands)
    self.assertIn("!MODE ENDPOINT", commands)
    self.assertIn("!ORDER COLUMN", commands)
    self.assertIn("!SPEED OFF", commands)

    readtype_call = next(
      c for c in self.send_command_mock.call_args_list if c.args[0] == "!READTYPE ABSPLA"
    )
    self.assertEqual(readtype_call.kwargs, {"num_res_fields": 2})

    mock_read_now.assert_called_once()
    mock_wait_for_idle.assert_called_once()
    mock_transfer_data.assert_called_once()

  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._wait_for_idle",
    new_callable=AsyncMock,
  )
  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._transfer_data",
    new_callable=AsyncMock,
    return_value="",
  )
  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._read_now",
    new_callable=AsyncMock,
  )
  async def test_read_fluorescence(self, mock_read_now, mock_transfer_data, mock_wait_for_idle):
    plate = AGenBio_96_wellplate_Ub_2200ul("test_plate")
    await self.backend.read_fluorescence(plate, [485], [520], [515])

    commands = [c.args[0] for c in self.send_command_mock.call_args_list]
    self.assertIn("!CLEAR DATA", commands)
    self.assertTrue(any(cmd.startswith("!XPOS") for cmd in commands))
    self.assertTrue(any(cmd.startswith("!YPOS") for cmd in commands))
    self.assertIn("!STRIP 1 12", commands)
    self.assertIn("!CSPEED 8", commands)
    self.assertIn("!SHAKE OFF", commands)
    self.assertIn("!FPW 10", commands)
    self.assertIn("!AUTOPMT ON", commands)
    self.assertIn("!EXWAVELENGTH 485", commands)
    self.assertIn("!EMWAVELENGTH 520", commands)
    self.assertIn("!AUTOFILTER OFF", commands)
    self.assertIn("!EMFILTER 515", commands)
    self.assertIn("!PMTCAL ONCE", commands)
    self.assertIn("!MODE ENDPOINT", commands)
    self.assertIn("!ORDER COLUMN", commands)
    self.assertIn("!READSTAGE TOP", commands)

    readtype_call = next(
      c for c in self.send_command_mock.call_args_list if c.args[0] == "!READTYPE FLU"
    )
    self.assertEqual(readtype_call.kwargs, {"num_res_fields": 1})

    mock_read_now.assert_called_once()
    mock_wait_for_idle.assert_called_once()
    mock_transfer_data.assert_called_once()

  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._wait_for_idle",
    new_callable=AsyncMock,
  )
  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._transfer_data",
    new_callable=AsyncMock,
    return_value="",
  )
  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._read_now",
    new_callable=AsyncMock,
  )
  async def test_read_luminescence(self, mock_read_now, mock_transfer_data, mock_wait_for_idle):
    plate = AGenBio_96_wellplate_Ub_2200ul("test_plate")
    await self.backend.read_luminescence(plate, [590])

    commands = [c.args[0] for c in self.send_command_mock.call_args_list]
    self.assertIn("!CLEAR DATA", commands)
    self.assertTrue(any(cmd.startswith("!XPOS") for cmd in commands))
    self.assertTrue(any(cmd.startswith("!YPOS") for cmd in commands))
    self.assertIn("!STRIP 1 12", commands)
    self.assertIn("!CSPEED 8", commands)
    self.assertIn("!SHAKE OFF", commands)
    self.assertIn("!EMWAVELENGTH 590", commands)
    self.assertIn("!PMTCAL ONCE", commands)
    self.assertIn("!MODE ENDPOINT", commands)
    self.assertIn("!ORDER COLUMN", commands)
    self.assertIn("!READSTAGE TOP", commands)

    readtype_call = next(
      c for c in self.send_command_mock.call_args_list if c.args[0] == "!READTYPE LUM"
    )
    self.assertEqual(readtype_call.kwargs, {"num_res_fields": 1})

    mock_read_now.assert_called_once()
    mock_wait_for_idle.assert_called_once()
    mock_transfer_data.assert_called_once()

  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._wait_for_idle",
    new_callable=AsyncMock,
  )
  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._transfer_data",
    new_callable=AsyncMock,
    return_value="",
  )
  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._read_now",
    new_callable=AsyncMock,
  )
  async def test_read_fluorescence_polarization(
    self,
    mock_read_now,
    mock_transfer_data,
    mock_wait_for_idle,
  ):
    plate = AGenBio_96_wellplate_Ub_2200ul("test_plate")
    await self.backend.read_fluorescence_polarization(plate, [485], [520], [515])

    commands = [c.args[0] for c in self.send_command_mock.call_args_list]
    self.assertIn("!CLEAR DATA", commands)
    self.assertTrue(any(cmd.startswith("!XPOS") for cmd in commands))
    self.assertTrue(any(cmd.startswith("!YPOS") for cmd in commands))
    self.assertIn("!STRIP 1 12", commands)
    self.assertIn("!CSPEED 8", commands)
    self.assertIn("!SHAKE OFF", commands)
    self.assertIn("!FPW 10", commands)
    self.assertIn("!AUTOPMT ON", commands)
    self.assertIn("!EXWAVELENGTH 485", commands)
    self.assertIn("!EMWAVELENGTH 520", commands)
    self.assertIn("!AUTOFILTER OFF", commands)
    self.assertIn("!EMFILTER 515", commands)
    self.assertIn("!PMTCAL ONCE", commands)
    self.assertIn("!MODE ENDPOINT", commands)
    self.assertIn("!ORDER COLUMN", commands)
    self.assertIn("!READSTAGE TOP", commands)

    readtype_call = next(
      c for c in self.send_command_mock.call_args_list if c.args[0] == "!READTYPE POLAR"
    )
    self.assertEqual(readtype_call.kwargs, {"num_res_fields": 1})

    mock_read_now.assert_called_once()
    mock_wait_for_idle.assert_called_once()
    mock_transfer_data.assert_called_once()

  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._wait_for_idle",
    new_callable=AsyncMock,
  )
  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._transfer_data",
    new_callable=AsyncMock,
    return_value="",
  )
  @patch(
    "pylabrobot.plate_reading.molecular_devices.backend.MolecularDevicesBackend._read_now",
    new_callable=AsyncMock,
  )
  async def test_read_time_resolved_fluorescence(
    self,
    mock_read_now,
    mock_transfer_data,
    mock_wait_for_idle,
  ):
    plate = AGenBio_96_wellplate_Ub_2200ul("test_plate")
    await self.backend.read_time_resolved_fluorescence(
      plate, [485], [520], [515], delay_time=10, integration_time=100
    )

    commands = [c.args[0] for c in self.send_command_mock.call_args_list]
    self.assertIn("!CLEAR DATA", commands)
    self.assertTrue(any(cmd.startswith("!XPOS") for cmd in commands))
    self.assertTrue(any(cmd.startswith("!YPOS") for cmd in commands))
    self.assertIn("!STRIP 1 12", commands)
    self.assertIn("!CSPEED 8", commands)
    self.assertIn("!SHAKE OFF", commands)
    self.assertIn("!FPW 50", commands)
    self.assertIn("!AUTOPMT ON", commands)
    self.assertIn("!EXWAVELENGTH 485", commands)
    self.assertIn("!EMWAVELENGTH 520", commands)
    self.assertIn("!AUTOFILTER OFF", commands)
    self.assertIn("!EMFILTER 515", commands)
    self.assertIn("!PMTCAL ONCE", commands)
    self.assertIn("!MODE ENDPOINT", commands)
    self.assertIn("!ORDER COLUMN", commands)
    self.assertIn("!COUNTTIMEDELAY 10", commands)
    self.assertIn("!COUNTTIME 0.1", commands)
    self.assertIn("!READSTAGE TOP", commands)

    readtype_call = next(
      c for c in self.send_command_mock.call_args_list if c.args[0] == "!READTYPE TIME 0 250"
    )
    self.assertEqual(readtype_call.kwargs, {"num_res_fields": 1})

    mock_read_now.assert_called_once()
    mock_wait_for_idle.assert_called_once()
    mock_transfer_data.assert_called_once()


class TestDataParsing(unittest.IsolatedAsyncioTestCase):
  send_command_mock: AsyncMock

  def setUp(self):
    with patch("pylabrobot.io.serial.Serial", return_value=MagicMock()):
      self.backend = MolecularDevicesBackend(port="COM1")
      self.send_command_mock = patch.object(
        self.backend, "send_command", new_callable=AsyncMock
      ).start()

  def test_parse_absorbance_single_wavelength(self):
    data_str = """
    12345.6	25.1	96-well
    L:	260
    L:	260
    1:	0.1	0.2
    2:	0.3	0.4
    """
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )

    result = self.backend._parse_data(data_str, settings)
    self.assertIsInstance(result, list)
    self.assertEqual(len(result), 1)
    read = result[0]
    self.assertEqual(read["wavelength"], 260)
    self.assertEqual(read["time"], 12345.6)
    self.assertEqual(read["temperature"], 25.1)
    self.assertEqual(read["data"], [[0.1, 0.3], [0.2, 0.4]])

  def test_parse_absorbance_multiple_wavelengths(self):
    data_str = """
    12345.6\t25.1\t96-well
    L:\t260\t280
    L:\t260
    1:\t0.1\t0.2
    2:\t0.3\t0.4
    L:\t280
    1:\t0.5\t0.6
    2:\t0.7\t0.8
    """
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    result = self.backend._parse_data(data_str, settings)
    self.assertIsInstance(result, list)
    self.assertEqual(len(result), 2)
    self.assertEqual(result[0]["wavelength"], 260)
    self.assertEqual(result[0]["data"], [[0.1, 0.3], [0.2, 0.4]])
    self.assertEqual(result[1]["wavelength"], 280)
    self.assertEqual(result[1]["data"], [[0.5, 0.7], [0.6, 0.8]])

  def test_parse_fluorescence(self):
    data_str = """
    12345.6\t25.1\t96-well
    exL:\t485
    emL:\t520
    L:\t485\t520
    1:\t100\t200
    2:\t300\t400
    """
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.FLU,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    result = self.backend._parse_data(data_str, settings)
    self.assertIsInstance(result, list)
    self.assertEqual(len(result), 1)
    read = result[0]
    self.assertEqual(read["ex_wavelength"], 485)
    self.assertEqual(read["em_wavelength"], 520)
    self.assertEqual(read["time"], 12345.6)
    self.assertEqual(read["temperature"], 25.1)
    self.assertEqual(read["data"], [[100.0, 300.0], [200.0, 400.0]])

  def test_parse_luminescence(self):
    data_str = """
    12345.6\t25.1\t96-well
    emL:\t590
    L:\t\t590
    1:\t1000\t2000
    2:\t3000\t4000
    """
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.LUM,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    result = self.backend._parse_data(data_str, settings)
    self.assertIsInstance(result, list)
    self.assertEqual(len(result), 1)
    read = result[0]
    self.assertEqual(read["em_wavelength"], 590)
    self.assertEqual(read["time"], 12345.6)
    self.assertEqual(read["temperature"], 25.1)
    self.assertEqual(read["data"], [[1000.0, 3000.0], [2000.0, 4000.0]])

  def test_parse_data_with_sat_and_nan(self):
    data_str = """
    12345.6\t25.1\t96-well
    L:\t260
    L:\t260
    1:\t0.1\t#SAT
    2:\t0.3\t-
    """
    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.ENDPOINT,
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
      spectrum_settings=None,
    )
    result = self.backend._parse_data(data_str, settings)
    self.assertIsInstance(result, list)
    self.assertEqual(len(result), 1)
    read = result[0]
    self.assertEqual(read["data"][1][0], float("inf"))
    self.assertTrue(math.isnan(read["data"][1][1]))

  async def test_parse_kinetic_absorbance(self):
    # Mock the send_command to return two different data blocks
    def data_generator():
      yield [
        "OK",
        """
    12345.6\t25.1\t96-well
    L:\t260
    L:\t260
    1:\t0.1\t0.2
    2:\t0.3\t0.4
    """,
      ]
      yield [
        "OK",
        """
    12355.6\t25.2\t96-well
    L:\t260
    L:\t260
    1:\t0.15\t0.25
    2:\t0.35\t0.45
    """,
      ]

    self.send_command_mock.side_effect = data_generator()

    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.KINETIC,
      kinetic_settings=KineticSettings(interval=10, num_readings=2),
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      spectrum_settings=None,
    )

    result = await self.backend._transfer_data(settings)
    self.assertEqual(len(result), 2)
    self.assertEqual(result[0]["wavelength"], 260)
    self.assertEqual(result[0]["data"], [[0.1, 0.3], [0.2, 0.4]])
    self.assertEqual(result[0]["time"], 12345.6)
    self.assertEqual(result[1]["wavelength"], 260)
    self.assertEqual(result[1]["data"], [[0.15, 0.35], [0.25, 0.45]])
    self.assertEqual(result[1]["time"], 12355.6)

  async def test_parse_spectrum_absorbance(self):
    # Mock the send_command to return two different data blocks for two wavelengths
    def data_generator():
      yield [
        "OK",
        """
    12345.6\t25.1\t96-well
    L:\t260
    L:\t260
    1:\t0.1\t0.2
    2:\t0.3\t0.4
    """,
      ]
      yield [
        "OK",
        """
    12355.6\t25.2\t96-well
    L:\t270
    L:\t270
    1:\t0.15\t0.25
    2:\t0.35\t0.45
    """,
      ]

    self.send_command_mock.side_effect = data_generator()

    settings = MolecularDevicesSettings(
      plate=MagicMock(),
      read_mode=ReadMode.ABS,
      read_type=ReadType.SPECTRUM,
      spectrum_settings=SpectrumSettings(start_wavelength=260, step=10, num_steps=2),
      read_order=ReadOrder.COLUMN,
      calibrate=Calibrate.ON,
      shake_settings=None,
      carriage_speed=CarriageSpeed.NORMAL,
      speed_read=False,
      kinetic_settings=None,
    )

    result = await self.backend._transfer_data(settings)
    self.assertEqual(len(result), 2)

    self.assertEqual(result[0]["wavelength"], 260)
    self.assertEqual(result[0]["data"], [[0.1, 0.3], [0.2, 0.4]])
    self.assertEqual(result[0]["time"], 12345.6)

    self.assertEqual(result[1]["wavelength"], 270)
    self.assertEqual(result[1]["data"], [[0.15, 0.35], [0.25, 0.45]])
    self.assertEqual(result[1]["time"], 12355.6)


class TestErrorHandling(unittest.IsolatedAsyncioTestCase):
  def setUp(self):
    self.mock_serial = MagicMock()
    self.mock_serial.setup = AsyncMock()
    self.mock_serial.stop = AsyncMock()
    self.mock_serial.write = AsyncMock()
    self.mock_serial.readline = AsyncMock()

    with patch("pylabrobot.io.serial.Serial", return_value=self.mock_serial):
      self.backend = MolecularDevicesBackend(port="/dev/tty01")
      self.backend.io = self.mock_serial

  async def _mock_send_command_response(self, response_str: str):
    self.mock_serial.readline.side_effect = [response_str.encode() + b">\r\n"]
    return await self.backend.send_command("!TEST")

  async def test_parse_basic_errors_fail_known_error_code(self):
    # Test a known error code (e.g., 107: no data to transfer)
    with self.assertRaisesRegex(
      MolecularDevicesUnrecognizedCommandError,
      "Command '!TEST' failed with error 107: no data to transfer",
    ):
      await self._mock_send_command_response("OK\t\r\n>FAIL\t 107")

  async def test_parse_basic_errors_fail_unknown_error_code(self):
    # Test an unknown error code
    with self.assertRaisesRegex(
      MolecularDevicesError, "Command '!TEST' failed with unknown error code: 999"
    ):
      await self._mock_send_command_response("FAIL\t 999")

  async def test_parse_basic_errors_fail_unparsable_error(self):
    # Test an unparsable error message (e.g., not an integer code)
    with self.assertRaisesRegex(
      MolecularDevicesError, "Command '!TEST' failed with unparsable error: FAIL\t ABC"
    ):
      await self._mock_send_command_response("FAIL\t ABC")

  async def test_parse_basic_errors_empty_response(self):
    # Test an empty response from the device
    self.mock_serial.readline.return_value = b""  # Simulate no response
    with self.assertRaisesRegex(TimeoutError, "Timeout waiting for response to command: !TEST"):
      await self.backend.send_command("!TEST", timeout=1)  # Short timeout for test

  async def test_parse_basic_errors_warning_response(self):
    # Test a response containing a warning
    self.mock_serial.readline.side_effect = [b"OK\tWarning: Something happened>\r\n"]
    # Expect no exception, but a warning logged (not directly testable with assertRaises)
    # We can assert that no error is raised.
    try:
      await self.backend.send_command("!TEST")
    except MolecularDevicesError:
      self.fail("MolecularDevicesError raised for a warning response")

  async def test_parse_basic_errors_ok_response(self):
    # Test a normal OK response
    self.mock_serial.readline.side_effect = [b"OK>\r\n"]
    try:
      response = await self.backend.send_command("!TEST")
      self.assertEqual(response, ["OK"])
    except MolecularDevicesError:
      self.fail("MolecularDevicesError raised for a valid OK response")


if __name__ == "__main__":
  unittest.main()



from dataclasses import dataclass

from pylabrobot.storage.cytomat.constants import (
  ActionRegister,
  ActionType,
  LoadStatusAtProcessor,
  LoadStatusFrontOfGate,
  OverviewRegister,
  SwapStationPosition,
)
from pylabrobot.storage.cytomat.utils import hex_to_binary


@dataclass(frozen=True)
class OverviewRegisterState:
  transfer_station_occupied: bool
  device_door_open: bool
  automatic_gate_open: bool
  handler_occupied: bool
  error_register_set: bool
  warning_register_set: bool
  ready_bit_set: bool
  busy_bit_set: bool

  @classmethod
  def from_resp(self, resp) -> "OverviewRegisterState":
    binary_value = hex_to_binary(resp)
    return OverviewRegisterState(
      **{member.name.lower(): binary_value[member.value] == "1" for member in OverviewRegister}
    )


@dataclass(frozen=True)
class ActionRegisterState:
  target: ActionType
  action: ActionRegister


@dataclass(frozen=True)
class SwapStationState:
  position: SwapStationPosition
  load_status_front_of_gate: LoadStatusFrontOfGate
  load_status_at_processor: LoadStatusAtProcessor


@dataclass(frozen=True)
class SensorStates:
  INIT_SENSOR_HEIGHT_MOTOR: bool
  INIT_SENSOR_CAROUSEL: bool
  SHOVEL_RETRACTED: bool
  SHOVEL_EXTENDED: bool
  SHOVEL_OCCUPIED: bool
  GATE_OPENED: bool
  GATE_CLOSED: bool
  TRANSFER_STATION_OCCUPIED: bool
  TRANSFER_STATION_POSITION_1: bool
  TRANSFER_STATION_POSITION_2: bool
  INNER_DOOR_OPENED: bool
  CAROUSEL_POSITION: bool
  HANDLER_POSITIONED_TOWARDS_STACKER: bool
  HANDLER_POSITIONED_TOWARDS_GATE: bool
  TRANSFER_STATION_SECOND_PLATE_OCCUPIED: bool



from typing import Optional

from pylabrobot.resources.coordinate import Coordinate
from pylabrobot.resources.resource import Resource
from pylabrobot.serializer import serialize


def get_child_location(resource: Resource) -> Coordinate:
  """
  If a resource is rotated, its rotated around its local origin. This does not always
  match up with the parent resource's origin. This function calculates the difference
  between the two origins and calculates the location of the resource correctly.
  """
  if not resource.rotation.y == resource.rotation.x == 0:
    raise ValueError("Resource rotation must be 0 around the x and y axes")
  if not resource.rotation.z % 90 == 0:
    raise ValueError("Resource rotation must be a multiple of 90 degrees on the z axis")
  location = {
    0.0: Coordinate(x=0, y=0, z=0),
    90.0: Coordinate(x=resource.get_size_y(), y=0, z=0),
    180.0: Coordinate(x=resource.get_size_x(), y=resource.get_size_y(), z=0),
    270.0: Coordinate(x=0, y=resource.get_size_x(), z=0),
  }[resource.rotation.z % 360]
  return location


class ResourceHolder(Resource):
  """
  A mixin class for resources that can hold other resources, like a plate or a lid.

  This applies a linear transformation after the rotation to correctly place the child resource.
  """

  def __init__(
    self,
    name,
    size_x,
    size_y,
    size_z,
    rotation=None,
    category="resource_holder",
    model=None,
    child_location: Coordinate = Coordinate.zero(),
    preferred_pickup_location: Optional[Coordinate] = None,
  ):
    Resource.__init__(
      self,
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      rotation=rotation,
      category=category,
      model=model,
      preferred_pickup_location=preferred_pickup_location,
    )
    self.child_location = child_location

  def get_default_child_location(self, resource: Resource) -> Coordinate:
    return get_child_location(resource) + self.child_location

  def assign_child_resource(
    self,
    resource: Resource,
    location: Optional[Coordinate] = None,
    reassign: bool = True,
  ):
    location = location or self.get_default_child_location(resource)
    if len(self.children) > 0 and not reassign:
      raise ValueError("ResourceHolders can only take one child at a time.")
    return super().assign_child_resource(resource, location, reassign)

  @property
  def resource(self) -> Optional[Resource]:
    if len(self.children) == 0:
      return None
    return self.children[0]

  @resource.setter
  def resource(self, resource: Optional[Resource]):
    if resource is None:
      if len(self.children) == 0:
        return
      self.unassign_child_resource(self.children[0])
      return
    self.assign_child_resource(resource, reassign=False)

  def serialize(self):
    return {**super().serialize(), "child_location": serialize(self.child_location)}

  def check_can_drop_resource_here(self, resource: Resource, *, reassign: bool = True) -> None:
    if self.resource is not None and resource is not self.resource:
      raise RuntimeError(
        f"Cannot drop resource {resource.name} onto resource holder {self.name} while it already has a resource. "
        "Please remove the resource before dropping a new one."
      )



import asyncio
import hashlib
import hmac
import logging
import re
import ssl
import warnings
import xml.etree.ElementTree as ET
from abc import ABCMeta
from base64 import b64decode
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, cast
from xml.dom import minidom

from pylabrobot.io import Socket
from pylabrobot.thermocycling.backend import ThermocyclerBackend
from pylabrobot.thermocycling.standard import LidStatus, Protocol, Stage, Step


def _generate_run_info_files(
  protocol: Protocol,
  block_id: int,
  sample_volume: float,
  run_mode: str,
  protocol_name: str,
  cover_enabled: bool,
  cover_temp: float,
  user_name: str,
  file_version="1.0.1",
  remote_run="true",
  hub="testhub",
  user="Guest",
  notes="",
  default_ramp_rate=100,
  ramp_rate_unit="DEGREES_PER_SECOND",
):
  root = ET.Element("TCProtocol")
  file_version_el = ET.SubElement(root, "FileVersion")
  file_version_el.text = file_version

  protocol_name_el = ET.SubElement(root, "ProtocolName")
  protocol_name_el.text = protocol_name

  user_name_el = ET.SubElement(root, "UserName")
  user_name_el.text = user_name

  block_id_el = ET.SubElement(root, "BlockID")
  block_id_el.text = str(block_id + 1)

  sample_volume_el = ET.SubElement(root, "SampleVolume")
  sample_volume_el.text = str(sample_volume)

  run_mode_el = ET.SubElement(root, "RunMode")
  run_mode_el.text = str(run_mode)

  cover_temp_el = ET.SubElement(root, "CoverTemperature")
  cover_temp_el.text = str(cover_temp)

  cover_setting_el = ET.SubElement(root, "CoverSetting")
  cover_setting_el.text = "On" if cover_enabled else "Off"

  for stage_obj in protocol.stages:
    if isinstance(stage_obj, Step):
      stage = Stage(steps=[stage_obj], repeats=1)
    else:
      stage = stage_obj

    stage_el = ET.SubElement(root, "TCStage")
    stage_flag_el = ET.SubElement(stage_el, "StageFlag")
    stage_flag_el.text = "CYCLING"

    num_repetitions_el = ET.SubElement(stage_el, "NumOfRepetitions")
    num_repetitions_el.text = str(stage.repeats)

    for step in stage.steps:
      step_el = ET.SubElement(stage_el, "TCStep")

      ramp_rate_el = ET.SubElement(step_el, "RampRate")
      ramp_rate_el.text = str(
        int(step.rate if step.rate is not None else default_ramp_rate) / 100 * 6
      )

      ramp_rate_unit_el = ET.SubElement(step_el, "RampRateUnit")
      ramp_rate_unit_el.text = ramp_rate_unit

      for t_val in step.temperature:
        temp_el = ET.SubElement(step_el, "Temperature")
        temp_el.text = str(t_val)

      hold_time_el = ET.SubElement(step_el, "HoldTime")
      if step.hold_seconds == float("inf"):
        hold_time_el.text = "-1"
      elif step.hold_seconds == 0:
        hold_time_el.text = "0"
      else:
        hold_time_el.text = str(step.hold_seconds)

      ext_temp_el = ET.SubElement(step_el, "ExtTemperature")
      ext_temp_el.text = "0"

      ext_hold_el = ET.SubElement(step_el, "ExtHoldTime")
      ext_hold_el.text = "0"

      ext_start_cycle_el = ET.SubElement(step_el, "ExtStartingCycle")
      ext_start_cycle_el.text = "1"

  rough_string = ET.tostring(root, encoding="utf-8")
  reparsed = minidom.parseString(rough_string)

  xml_declaration = '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n'
  pretty_xml_as_string = (
    xml_declaration + reparsed.toprettyxml(indent="  ")[len('<?xml version="1.0" ?>') :]
  )

  output2_lines = [
    f"-remoterun= {remote_run}",
    f"-hub= {hub}",
    f"-user= {user}",
    f"-method= {protocol_name}",
    f"-volume= {sample_volume}",
    f"-cover= {cover_temp}",
    f"-mode= {run_mode}",
    f"-coverEnabled= {'On' if cover_enabled else 'Off'}",
    f"-notes= {notes}",
  ]
  output2_string = "\n".join(output2_lines)

  return pretty_xml_as_string, output2_string


def _gen_protocol_data(
  protocol: Protocol,
  block_id: int,
  sample_volume: float,
  run_mode: str,
  cover_temp: float,
  cover_enabled: bool,
  protocol_name: str,
  stage_name_prefixes: List[str],
):
  def step_to_scpi(step: Step, step_index: int) -> dict:
    multiline: List[dict] = []

    infinite_hold = step.hold_seconds == float("inf")

    if infinite_hold and min(step.temperature) < 20:
      multiline.append({"cmd": "CoverRAMP", "params": {}, "args": ["30"]})

    multiline.append(
      {
        "cmd": "RAMP",
        "params": {"rate": str(step.rate if step.rate is not None else 100)},
        "args": [str(t) for t in step.temperature],
      }
    )

    if infinite_hold:
      multiline.append({"cmd": "HOLD", "params": {}, "args": []})
    elif step.hold_seconds > 0:
      multiline.append({"cmd": "HOLD", "params": {}, "args": [str(step.hold_seconds)]})

    return {
      "cmd": "STEP",
      "params": {},
      "args": [str(step_index)],
      "tag": "multiline.step",
      "multiline": multiline,
    }

  def stage_to_scpi(stage: Stage, stage_index: int, stage_name_prefix: str) -> dict:
    return {
      "cmd": "STAGe",
      "params": {"repeat": str(stage.repeats)},
      "args": [stage_index, f"{stage_name_prefix}_{stage_index}"],
      "tag": "multiline.stage",
      "multiline": [step_to_scpi(step, i + 1) for i, step in enumerate(stage.steps)],
    }

  stages = protocol.stages
  assert len(stages) == len(stage_name_prefixes), (
    "Number of stages must match number of stage names"
  )

  data = {
    # "status": "OK",
    "cmd": f"TBC{block_id + 1}:Protocol",
    "params": {"Volume": str(sample_volume), "RunMode": run_mode},
    "args": [protocol_name],
    "tag": "multiline.outer",
    "multiline": [
      stage_to_scpi(stage, stage_index=i + 1, stage_name_prefix=stage_name_prefix)
      for i, (stage, stage_name_prefix) in enumerate(zip(stages, stage_name_prefixes))
    ],
    "_blockId": block_id + 1,
    "_coverTemp": cover_temp,
    "_coverEnabled": "On" if cover_enabled else "Off",
    "_infinite_holds": [
      [stage_index, step_index]
      for stage_index, stage in enumerate(stages)
      for step_index, step in enumerate(stage.steps)
      if step.hold_seconds == float("inf")
    ],
  }

  return data


class ThermoFisherThermocyclerBackend(ThermocyclerBackend, metaclass=ABCMeta):
  """Backend for Proflex thermocycler."""

  def __init__(
    self,
    ip: str,
    use_ssl: bool = False,
    serial_number: Optional[str] = None,
    port: Optional[int] = None,
  ):
    if port is not None:
      raise NotImplementedError("Specifying a port is deprecated. Use use_ssl instead.")

    self.ip = ip
    self.use_ssl = use_ssl

    if use_ssl:
      self.port = 7443
      if serial_number is None:
        raise ValueError("Serial number is required for SSL connection (port 7443)")
      self.device_shared_secret = f"53rv1c3{serial_number}".encode("utf-8")

      ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
      ssl_context.check_hostname = False
      ssl_context.verify_mode = ssl.CERT_NONE
      # TLSv1 is required for legacy ThermoFisher hardware - silence deprecation warning
      with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="ssl.TLSVersion.TLSv1 is deprecated")
        ssl_context.minimum_version = ssl.TLSVersion.TLSv1
        ssl_context.maximum_version = ssl.TLSVersion.TLSv1
      try:
        # This is required for some legacy devices that use older ciphers or protocols
        # that are disabled by default in newer OpenSSL versions.
        ssl_context.set_ciphers("DEFAULT:@SECLEVEL=0")
      except (ValueError, ssl.SSLError):
        # This might fail on some systems/implementations, but it's worth a try
        pass
    else:
      self.port = 7000
      self.device_shared_secret = b"f4ct0rymt55"
      ssl_context = None

    self.io = Socket(
      human_readable_device_name="Thermo Fisher Thermocycler",
      host=ip,
      port=self.port,
      ssl_context=ssl_context,
      server_hostname=serial_number,
    )
    self._num_blocks: Optional[int] = None
    self.num_temp_zones = 0
    self.available_blocks: List[int] = []
    self.logger = logging.getLogger("pylabrobot.thermocycling.proflex")
    self.current_runs: Dict[int, str] = {}

  @property
  def num_blocks(self) -> int:
    if self._num_blocks is None:
      raise ValueError("Number of blocks not set. Call setup() first.")
    return self._num_blocks

  def _get_auth_token(self, challenge: str):
    challenge_bytes = challenge.encode("utf-8")
    return hmac.new(self.device_shared_secret, challenge_bytes, hashlib.md5).hexdigest()

  def _build_scpi_msg(self, data: dict) -> str:
    def generate_output(data_dict: dict, indent_level=0) -> str:
      lines = []
      if indent_level == 0:
        line = data_dict["cmd"]
        for k, v in data_dict.get("params", {}).items():
          if v is True:
            line += f" -{k}"
          elif v is False:
            pass
          else:
            line += f" -{k}={v}"
        for val in data_dict.get("args", []):
          line += f" {val}"
        if "multiline" in data_dict:
          line += f" <{data_dict['tag']}>"
        lines.append(line)

        if "multiline" in data_dict:
          lines += generate_multiline(data_dict, indent_level + 1)
          lines.append(f"</{data_dict['tag']}>")
      return "\n".join(lines)

    def generate_multiline(multi_dict, indent_level=0) -> List[str]:
      def indent():
        return " " * 8 * indent_level

      lines = []
      for element in multi_dict["multiline"]:
        line = indent() + element["cmd"]
        for k, v in element.get("params", {}).items():
          line += f" -{k}={v}"
        for arg in element.get("args", []):
          line += f" {arg}"

        if "multiline" in element:
          line += f" <{element['tag']}>"
          lines.append(line)
          lines += generate_multiline(element, indent_level + 1)
          lines.append(indent() + f"</{element['tag']}>")
        else:
          lines.append(line)
      return lines

    return generate_output(data) + "\r\n"

  def _parse_scpi_response(self, response: str):
    START_TAG_REGEX = re.compile(r"(.*?)<(multiline\.[a-zA-Z0-9_]+)>")
    END_TAG_REGEX = re.compile(r"</(multiline\.[a-zA-Z0-9_]+)>")
    PARAM_REGEX = re.compile(r"^-([A-Za-z0-9_]+)(?:=(.*))?$")

    def parse_command_line(line):
      start_match = START_TAG_REGEX.search(line)
      if start_match:
        cmd_part = start_match.group(1).strip()
        tag_name = start_match.group(2)
      else:
        cmd_part = line
        tag_name = None

      if not cmd_part:
        return None, [], tag_name

      parts = cmd_part.split()
      command = parts[0]
      args = parts[1:]
      return command, args, tag_name

    def process_args(args_list):
      params: Dict[str, Any] = {}
      positional_args = []
      for arg in args_list:
        match = PARAM_REGEX.match(arg)
        if match:
          key = match.group(1)
          value = match.group(2)
          if value is None:
            params[key] = True
          else:
            params[key] = value
        else:
          positional_args.append(arg)
      return positional_args, params

    def parse_structure(scpi_resp: str):
      first_space_idx = scpi_resp.find(" ")
      status = scpi_resp[:first_space_idx]
      scpi_resp = scpi_resp[first_space_idx + 1 :]
      lines = scpi_resp.split("\n")

      root = {"status": status, "multiline": []}
      stack = [root]

      for original_line in lines:
        line = original_line.strip()
        if not line:
          continue
        end_match = END_TAG_REGEX.match(line)
        if end_match:
          if len(stack) > 1:
            stack.pop()
          else:
            raise ValueError("Unmatched end tag: </{}>".format(end_match.group(1)))
          continue

        command, args, start_tag = parse_command_line(line)
        if command is not None:
          pos_args, params = process_args(args)
          node = {"cmd": command, "args": pos_args, "params": params}
          if start_tag:
            node["multiline"] = []
            stack[-1]["multiline"].append(node)  # type: ignore
            stack.append(node)
            node["tag"] = start_tag
          else:
            stack[-1]["multiline"].append(node)  # type: ignore

      if len(stack) != 1:
        raise ValueError("Unbalanced tags in response.")
      return root

    if response.startswith("ERRor"):
      raise ValueError(f"Error response: {response}")

    result = parse_structure(response)
    status_val = result["status"]
    result = result["multiline"][0]
    result["status"] = status_val
    return result

  async def _read_response(self, timeout=1, read_once=True) -> str:
    try:
      if read_once:
        response_b = await self.io.read(timeout=timeout)
      else:
        response_b = await self.io.read_until_eof(timeout=timeout)
      response = response_b.decode("ascii")
      self.logger.debug("Response received: %s", response)
      return response
    except TimeoutError:
      return ""
    except Exception as e:
      self.logger.error("Error reading from socket: %s", e)
      return ""

  async def send_command(self, data, response_timeout=1, read_once=True):
    msg = self._build_scpi_msg(data)
    self.logger.debug("Command sent: %s", msg.strip())

    await self.io.write(msg.encode("ascii"), timeout=response_timeout)
    return await self._read_response(timeout=response_timeout, read_once=read_once)

  async def _scpi_authenticate(self):
    await self.io.setup()
    await self._read_response(timeout=5)
    challenge_res = await self.send_command({"cmd": "CHAL?"})
    challenge = self._parse_scpi_response(challenge_res)["args"][0]
    auth = self._get_auth_token(challenge)
    auth_res = await self.send_command({"cmd": "AUTH", "args": [auth]})
    if self._parse_scpi_response(auth_res)["status"] != "OK":
      raise ValueError("Authentication failed")
    acc_res = await self.send_command(
      {"cmd": "ACCess", "params": {"stealth": True}, "args": ["Controller"]}
    )
    if self._parse_scpi_response(acc_res)["status"] != "OK":
      raise ValueError("Access failed")

  async def _load_num_blocks_and_type(self):
    block_present_val = await self.get_block_presence()
    if block_present_val == "0":
      raise ValueError("Block not present")
    self.bid = await self.get_block_id()
    if self.bid == "12":
      self._num_blocks = 1
      self.num_temp_zones = 6
    elif self.bid == "13":
      self._num_blocks = 3
      self.num_temp_zones = 2
    elif self.bid == "31":
      self._num_blocks = 1
      self.num_temp_zones = 3
    else:
      raise NotImplementedError("Only BID 31, 12 and 13 are supported")

  async def is_block_running(self, block_id: int) -> bool:
    run_name = await self.get_run_name(block_id=block_id)
    return run_name != "-"

  async def _load_available_blocks(self) -> None:
    await (
      self._scpi_authenticate()
    )  # in case users wants to see available blocks without setting them up
    await self._load_num_blocks_and_type()
    assert self._num_blocks is not None, "Number of blocks not set"
    for block_id in range(self._num_blocks):
      block_error = await self.get_error(block_id=block_id)
      if block_error != "0":
        raise ValueError(f"Block {block_id} has error: {block_error}")
      if not await self.is_block_running(block_id=block_id):
        if block_id not in self.available_blocks:
          self.available_blocks.append(block_id)

  async def get_block_current_temperature(self, block_id=1) -> List[float]:
    res = await self.send_command({"cmd": f"TBC{block_id + 1}:TBC:BlockTemperatures?"})
    return cast(List[float], self._parse_scpi_response(res)["args"])

  async def get_sample_temps(self, block_id=1) -> List[float]:
    res = await self.send_command({"cmd": f"TBC{block_id + 1}:TBC:SampleTemperatures?"})
    return cast(List[float], self._parse_scpi_response(res)["args"])

  async def get_nickname(self) -> str:
    res = await self.send_command({"cmd": "SYST:SETT:NICK?"})
    return cast(str, self._parse_scpi_response(res)["args"][0])

  async def set_nickname(self, nickname: str) -> None:
    res = await self.send_command({"cmd": "SYST:SETT:NICK", "args": [nickname]})
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to set nickname")

  async def get_log_by_runname(self, run_name: str) -> str:
    res = await self.send_command(
      {"cmd": "FILe:READ?", "args": [f"RUNS:{run_name}/{run_name}.log"]},
      response_timeout=5,
      read_once=False,
    )
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to get log")
    res.replace("\n", "")
    # Extract the base64 encoded log content between <quote> tags
    encoded_log_match = re.search(r"<quote>(.*?)</quote>", res, re.DOTALL)
    if not encoded_log_match:
      raise ValueError("Failed to parse log content")
    encoded_log = encoded_log_match.group(1).strip()
    log = b64decode(encoded_log).decode("utf-8")
    return log

  async def get_elapsed_run_time_from_log(self, run_name: str) -> int:
    """
    Parses a log to find the elapsed run time in hh:mm:ss format
    and converts it to total seconds.
    """
    log = await self.get_log_by_runname(run_name)

    # Updated regex to capture hours, minutes, and seconds
    elapsed_time_match = re.search(r"Run Time:\s*(\d+):(\d+):(\d+)", log)

    if not elapsed_time_match:
      raise ValueError("Failed to parse elapsed time from log. Expected hh:mm:ss format.")

    # Extract h, m, s, and convert them to integers
    hours = int(elapsed_time_match.group(1))
    minutes = int(elapsed_time_match.group(2))
    seconds = int(elapsed_time_match.group(3))

    # Calculate the total seconds
    total_seconds = (hours * 3600) + (minutes * 60) + seconds

    return total_seconds

  async def set_block_idle_temp(
    self, temp: float, block_id: int, control_enabled: bool = True
  ) -> None:
    if block_id not in self.available_blocks:
      raise ValueError(f"Block {block_id} is not available")
    res = await self.send_command(
      {"cmd": f"TBC{block_id + 1}:BLOCK", "args": [1 if control_enabled else 0, temp]}
    )
    if self._parse_scpi_response(res)["status"] != "NEXT":
      raise ValueError("Failed to set block idle temperature")
    follow_up = await self._read_response()
    if self._parse_scpi_response(follow_up)["status"] != "OK":
      raise ValueError("Failed to set block idle temperature")

  async def set_cover_idle_temp(
    self, temp: float, block_id: int, control_enabled: bool = True
  ) -> None:
    if block_id not in self.available_blocks:
      raise ValueError(f"Block {block_id} not available")
    res = await self.send_command(
      {"cmd": f"TBC{block_id + 1}:COVER", "args": [1 if control_enabled else 0, temp]}
    )
    if self._parse_scpi_response(res)["status"] != "NEXT":
      raise ValueError("Failed to set cover idle temperature")
    follow_up = await self._read_response()
    if self._parse_scpi_response(follow_up)["status"] != "OK":
      raise ValueError("Failed to set cover idle temperature")

  async def set_block_temperature(
    self, temperature: List[float], block_id: Optional[int] = None, rate: float = 100
  ):
    if block_id not in self.available_blocks:
      raise ValueError(f"Block {block_id} not available")
    res = await self.send_command(
      {"cmd": f"TBC{block_id + 1}:RAMP", "params": {"rate": rate}, "args": temperature},
      response_timeout=60,
    )
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to ramp block temperature")

  async def block_ramp_single_temp(self, target_temp: float, block_id: int, rate: float = 100):
    """Set a single temperature for the block with a ramp rate.

    It might be better to use `set_block_temperature` to set individual temperatures for each zone.
    """
    if block_id not in self.available_blocks:
      raise ValueError(f"Block {block_id} not available")
    res = await self.send_command(
      {"cmd": f"TBC{block_id + 1}:BlockRAMP", "params": {"rate": rate}, "args": [target_temp]},
      response_timeout=60,
    )
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to ramp block temperature")

  async def set_lid_temperature(self, temperature: List[float], block_id: Optional[int] = None):
    assert block_id is not None, "block_id must be specified"
    assert len(set(temperature)) == 1, "Lid temperature must be the same for all zones"
    target_temp = temperature[0]
    if block_id not in self.available_blocks:
      raise ValueError(f"Block {block_id} not available")
    res = await self.send_command(
      {"cmd": f"TBC{block_id + 1}:CoverRAMP", "params": {}, "args": [target_temp]},
      response_timeout=60,
    )
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to ramp cover temperature")

  async def buzzer_on(self):
    res = await self.send_command({"cmd": "BUZZer+"})
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to turn on buzzer")

  async def buzzer_off(self):
    res = await self.send_command({"cmd": "BUZZer-"})
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to turn off buzzer")

  async def send_morse_code(self, morse_code: str):
    short_beep_duration = 0.1
    long_beep_duration = short_beep_duration * 3
    space_duration = short_beep_duration * 3
    assert all(char in ".- " for char in morse_code), "Invalid characters in morse code"
    for char in morse_code:
      if char == ".":
        await self.buzzer_on()
        await asyncio.sleep(short_beep_duration)
        await self.buzzer_off()
      elif char == "-":
        await self.buzzer_on()
        await asyncio.sleep(long_beep_duration)
        await self.buzzer_off()
      elif char == " ":
        await asyncio.sleep(space_duration)
      await asyncio.sleep(short_beep_duration)  # between letters is a short unit

  async def continue_run(self, block_id: int):
    for _ in range(3):
      await asyncio.sleep(1)
      res = await self.send_command({"cmd": f"TBC{block_id + 1}:CONTinue"})
      if self._parse_scpi_response(res)["status"] != "OK":
        raise ValueError("Failed to continue from indefinite hold")

  async def _write_file(self, filename: str, data: str, encoding="plain"):
    write_res = await self.send_command(
      {
        "cmd": "FILe:WRITe",
        "params": {"encoding": encoding},
        "args": [filename],
        "multiline": [{"cmd": data}],
        "tag": "multiline.write",
      },
      response_timeout=1,
      read_once=False,
    )
    if self._parse_scpi_response(write_res)["status"] != "OK":
      raise ValueError("Failed to write file")

  async def get_block_id(self):
    res = await self.send_command({"cmd": "TBC:BID?"})
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to get block ID")
    return self._parse_scpi_response(res)["args"][0]

  async def get_block_presence(self):
    res = await self.send_command({"cmd": "TBC:BlockPresence?"})
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to get block presence")
    return self._parse_scpi_response(res)["args"][0]

  async def check_run_exists(self, run_name: str) -> bool:
    res = await self.send_command(
      {"cmd": "RUNS:EXISTS?", "args": [run_name], "params": {"type": "folders"}}
    )
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to check if run exists")
    return cast(str, self._parse_scpi_response(res)["args"][1]) == "True"

  async def create_run(self, run_name: str):
    res = await self.send_command({"cmd": "RUNS:NEW", "args": [run_name]}, response_timeout=10)
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to create run")
    return self._parse_scpi_response(res)["args"][0]

  async def get_run_name(self, block_id: int) -> str:
    res = await self.send_command({"cmd": f"TBC{block_id + 1}:RUNTitle?"})
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to get run title")
    return cast(str, self._parse_scpi_response(res)["args"][0])

  async def _get_run_progress(self, block_id: int):
    res = await self.send_command({"cmd": f"TBC{block_id + 1}:RUNProgress?"})
    parsed_res = self._parse_scpi_response(res)
    if parsed_res["status"] != "OK":
      raise ValueError("Failed to get run status")
    if parsed_res["cmd"] == f"TBC{block_id + 1}:RunProtocol":
      await self._read_response()
      return False
    return self._parse_scpi_response(res)["params"]

  async def get_estimated_run_time(self, block_id: int):
    res = await self.send_command({"cmd": f"TBC{block_id + 1}:ESTimatedTime?"})
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to get estimated run time")
    return self._parse_scpi_response(res)["args"][0]

  async def get_elapsed_run_time(self, block_id: int):
    res = await self.send_command({"cmd": f"TBC{block_id + 1}:ELAPsedTime?"})
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to get elapsed run time")
    return int(self._parse_scpi_response(res)["args"][0])

  async def get_remaining_run_time(self, block_id):
    res = await self.send_command({"cmd": f"TBC{block_id + 1}:REMainingTime?"})
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to get remaining run time")
    return int(self._parse_scpi_response(res)["args"][0])

  async def get_error(self, block_id):
    res = await self.send_command({"cmd": f"TBC{block_id + 1}:ERROR?"})
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to get error")
    return self._parse_scpi_response(res)["args"][0]

  async def power_on(self):
    res = await self.send_command({"cmd": "POWER", "args": ["On"]}, response_timeout=20)
    if res == "" or self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to power on")

  async def power_off(self):
    res = await self.send_command({"cmd": "POWER", "args": ["Off"]})
    if self._parse_scpi_response(res)["status"] != "OK":
      raise ValueError("Failed to power off")

  async def _scpi_write_run_info(
    self,
    protocol: Protocol,
    run_name: str,
    block_id: int,
    sample_volume: float,
    run_mode: str,
    protocol_name: str,
    cover_temp: float,
    cover_enabled: bool,
    user_name: str,
  ):
    xmlfile, tmpfile = _generate_run_info_files(
      protocol=protocol,
      block_id=block_id,
      sample_volume=sample_volume,
      run_mode=run_mode,
      protocol_name=protocol_name,
      cover_temp=cover_temp,
      cover_enabled=cover_enabled,
      user_name="LifeTechnologies",  # for some reason LifeTechnologies is used here
    )
    await self._write_file(f"runs:{run_name}/{protocol_name}.method", xmlfile)
    await self._write_file(f"runs:{run_name}/{run_name}.tmp", tmpfile)

  async def _scpi_run_protocol(
    self,
    protocol: Protocol,
    run_name: str,
    block_id: int,
    sample_volume: float,
    run_mode: str,
    protocol_name: str,
    cover_temp: float,
    cover_enabled: bool,
    user_name: str,
    stage_name_prefixes: List[str],
  ):
    load_res = await self.send_command(
      data=_gen_protocol_data(
        protocol=protocol,
        block_id=block_id,
        sample_volume=sample_volume,
        run_mode=run_mode,
        cover_temp=cover_temp,
        cover_enabled=cover_enabled,
        protocol_name=protocol_name,
        stage_name_prefixes=stage_name_prefixes,
      ),
      response_timeout=5,
      read_once=False,
    )
    if self._parse_scpi_response(load_res)["status"] != "OK":
      self.logger.error(load_res)
      self.logger.error("Protocol failed to load")
      raise ValueError("Protocol failed to load")

    start_res = await self.send_command(
      {
        "cmd": f"TBC{block_id + 1}:RunProtocol",
        "params": {
          "User": user_name,
          "CoverTemperature": cover_temp,
          "CoverEnabled": "On" if cover_enabled else "Off",
        },
        "args": [protocol_name, run_name],
      },
      response_timeout=2,
      read_once=False,
    )

    if self._parse_scpi_response(start_res)["status"] == "NEXT":
      self.logger.info("Protocol started")
    else:
      self.logger.error(start_res)
      self.logger.error("Protocol failed to start")
      raise ValueError("Protocol failed to start")

    total_time = await self.get_estimated_run_time(block_id=block_id)
    total_time = float(total_time)
    self.logger.info(f"Estimated run time: {total_time}")
    self.current_runs[block_id] = run_name

  async def abort_run(self, block_id: int):
    if not await self.is_block_running(block_id=block_id):
      self.logger.info("Failed to abort protocol: no run is currently running on this block")
      return
    run_name = await self.get_run_name(block_id=block_id)
    abort_res = await self.send_command({"cmd": f"TBC{block_id + 1}:AbortRun", "args": [run_name]})
    if self._parse_scpi_response(abort_res)["status"] != "OK":
      self.logger.error(abort_res)
      self.logger.error("Failed to abort protocol")
      raise ValueError("Failed to abort protocol")
    self.logger.info("Protocol aborted")
    await asyncio.sleep(10)

  @dataclass
  class RunProgress:
    stage: str
    elapsed_time: int
    remaining_time: int
    running: bool

  async def get_run_info(self, protocol: Protocol, block_id: int) -> "RunProgress":
    progress = await self._get_run_progress(block_id=block_id)
    run_name = await self.get_run_name(block_id=block_id)
    if not progress:
      self.logger.info("Protocol completed")
      return ThermoFisherThermocyclerBackend.RunProgress(
        running=False,
        stage="completed",
        elapsed_time=await self.get_elapsed_run_time_from_log(run_name=run_name),
        remaining_time=0,
      )

    if progress["RunTitle"] == "-":
      await self._read_response(timeout=5)
      self.logger.info("Protocol completed")
      return ThermoFisherThermocyclerBackend.RunProgress(
        running=False,
        stage="completed",
        elapsed_time=await self.get_elapsed_run_time_from_log(run_name=run_name),
        remaining_time=0,
      )

    if progress["Stage"] == "POSTRun":
      self.logger.info("Protocol in POSTRun")
      return ThermoFisherThermocyclerBackend.RunProgress(
        running=True,
        stage="POSTRun",
        elapsed_time=await self.get_elapsed_run_time_from_log(run_name=run_name),
        remaining_time=0,
      )

    # TODO: move to separate wait method
    time_elapsed = await self.get_elapsed_run_time(block_id=block_id)
    remaining_time = await self.get_remaining_run_time(block_id=block_id)

    if progress["Stage"] != "-" and progress["Step"] != "-":
      current_step = protocol.stages[int(progress["Stage"]) - 1].steps[int(progress["Step"]) - 1]
      if current_step.hold_seconds == float("inf"):
        while True:
          block_temps = await self.get_block_current_temperature(block_id=block_id)
          target_temps = current_step.temperature
          if all(
            abs(float(block_temps[i]) - target_temps[i]) < 0.5 for i in range(len(block_temps))
          ):
            break
          await asyncio.sleep(5)
        self.logger.info("Infinite hold")
        return ThermoFisherThermocyclerBackend.RunProgress(
          running=False,
          stage="infinite_hold",
          elapsed_time=time_elapsed,
          remaining_time=remaining_time,
        )

    self.logger.info(f"Elapsed time: {time_elapsed}")
    self.logger.info(f"Remaining time: {remaining_time}")
    return ThermoFisherThermocyclerBackend.RunProgress(
      running=True,
      stage=progress["Stage"],
      elapsed_time=time_elapsed,
      remaining_time=remaining_time,
    )

  # *************Methods implementing ThermocyclerBackend***********************

  async def setup(
    self, block_idle_temp=25, cover_idle_temp=105, blocks_to_setup: Optional[List[int]] = None
  ):
    await self._scpi_authenticate()
    await self.power_on()
    await self._load_num_blocks_and_type()
    if blocks_to_setup is None:
      await self._load_available_blocks()
      if len(self.available_blocks) == 0:
        raise ValueError("No available blocks. Set blocks_to_setup to force setup")
    else:
      self.available_blocks = blocks_to_setup
    for block_index in self.available_blocks:
      await self.set_block_idle_temp(temp=block_idle_temp, block_id=block_index)
      await self.set_cover_idle_temp(temp=cover_idle_temp, block_id=block_index)

  async def deactivate_lid(self, block_id: Optional[int] = None):
    assert block_id is not None, "block_id must be specified"
    return await self.set_cover_idle_temp(temp=105, control_enabled=False, block_id=block_id)

  async def deactivate_block(self, block_id: Optional[int] = None):
    assert block_id is not None, "block_id must be specified"
    return await self.set_block_idle_temp(temp=25, control_enabled=False, block_id=block_id)

  async def get_lid_current_temperature(self, block_id: Optional[int] = None) -> List[float]:
    assert block_id is not None, "block_id must be specified"
    res = await self.send_command({"cmd": f"TBC{block_id + 1}:TBC:CoverTemperatures?"})
    return cast(List[float], self._parse_scpi_response(res)["args"])

  async def run_protocol(
    self,
    protocol: Protocol,
    block_max_volume: float,
    block_id: Optional[int] = None,
    run_name="testrun",
    user="Admin",
    run_mode: str = "Fast",
    cover_temp: float = 105,
    cover_enabled=True,
    protocol_name: str = "PCR_Protocol",
    stage_name_prefixes: Optional[List[str]] = None,
  ):
    assert block_id is not None, "block_id must be specified"

    if await self.check_run_exists(run_name):
      self.logger.warning(f"Run {run_name} already exists")
    else:
      await self.create_run(run_name)

    # wrap all Steps in Stage objects where necessary
    for i, stage in enumerate(protocol.stages):
      if isinstance(stage, Step):
        protocol.stages[i] = Stage(steps=[stage], repeats=1)

    for stage in protocol.stages:
      for step in stage.steps:
        if len(step.temperature) != self.num_temp_zones:
          raise ValueError(
            f"Each step in the protocol must have a list of temperatures "
            f"of length {self.num_temp_zones}. "
            f"Step temperatures: {step.temperature} (length {len(step.temperature)})"
          )

    stage_name_prefixes = stage_name_prefixes or ["Stage_" for i in range(len(protocol.stages))]

    await self._scpi_write_run_info(
      protocol=protocol,
      block_id=block_id,
      run_name=run_name,
      user_name=user,
      sample_volume=block_max_volume,
      run_mode=run_mode,
      cover_temp=cover_temp,
      cover_enabled=cover_enabled,
      protocol_name=protocol_name,
    )
    await self._scpi_run_protocol(
      protocol=protocol,
      run_name=run_name,
      block_id=block_id,
      sample_volume=block_max_volume,
      run_mode=run_mode,
      cover_temp=cover_temp,
      cover_enabled=cover_enabled,
      protocol_name=protocol_name,
      user_name=user,
      stage_name_prefixes=stage_name_prefixes,
    )

  async def stop(self):
    for block_id in self.current_runs.keys():
      await self.abort_run(block_id=block_id)

      await self.deactivate_lid(block_id=block_id)
      await self.deactivate_block(block_id=block_id)

    await self.io.stop()

  async def get_block_status(self, *args, **kwargs):
    raise NotImplementedError

  async def get_current_cycle_index(self, block_id: Optional[int] = None) -> int:
    assert block_id is not None, "block_id must be specified"
    progress = await self._get_run_progress(block_id=block_id)
    if progress is None:
      raise RuntimeError("No progress information available")

    if progress["RunTitle"] == "-":
      await self._read_response(timeout=5)
      raise RuntimeError("Protocol completed or not started")

    if progress["Stage"] == "POSTRun":
      raise RuntimeError("Protocol in POSTRun stage, no current cycle index")

    if progress["Stage"] != "-" and progress["Step"] != "-":
      return int(progress["Stage"]) - 1

    raise RuntimeError("Current cycle index is not available, protocol may not be running")

  async def get_current_step_index(self, block_id: Optional[int] = None) -> int:
    assert block_id is not None, "block_id must be specified"
    progress = await self._get_run_progress(block_id=block_id)
    if progress is None:
      raise RuntimeError("No progress information available")

    if progress["RunTitle"] == "-":
      await self._read_response(timeout=5)
      raise RuntimeError("Protocol completed or not started")

    if progress["Stage"] == "POSTRun":
      raise RuntimeError("Protocol in POSTRun stage, no current cycle index")

    if progress["Stage"] != "-" and progress["Step"] != "-":
      return int(progress["Step"]) - 1

    raise RuntimeError("Current step index is not available, protocol may not be running")

  async def get_hold_time(self, *args, **kwargs):
    # deprecated
    raise NotImplementedError

  async def get_lid_open(self, *args, **kwargs):
    raise NotImplementedError("Proflex thermocycler does not support lid open status check")

  async def get_lid_status(self, *args, **kwargs) -> LidStatus:
    raise NotImplementedError

  async def get_lid_target_temperature(self, *args, **kwargs):
    # deprecated
    raise NotImplementedError

  async def get_total_cycle_count(self, *args, **kwargs):
    # deprecated
    raise NotImplementedError

  async def get_total_step_count(self, *args, **kwargs):
    # deprecated
    raise NotImplementedError

  async def get_block_target_temperature(self, *args, **kwargs):
    # deprecated
    raise NotImplementedError



import asyncio
import logging
import time
import warnings
from typing import List, Literal, Optional, Union, cast

try:
  import serial

  HAS_SERIAL = True
except ImportError as e:
  HAS_SERIAL = False
  _SERIAL_IMPORT_ERROR = e

from pylabrobot.io.serial import Serial
from pylabrobot.resources import Plate, PlateCarrier, PlateHolder
from pylabrobot.storage.backend import IncubatorBackend
from pylabrobot.storage.cytomat.constants import (
  ActionRegister,
  ActionType,
  CytomatActionResponse,
  CytomatIncupationResponse,
  CytomatType,
  ErrorRegister,
  LoadStatusAtProcessor,
  LoadStatusFrontOfGate,
  SensorRegister,
  SwapStationPosition,
  WarningRegister,
)
from pylabrobot.storage.cytomat.errors import (
  CytomatBusyError,
  CytomatCommandUnknownError,
  CytomatTelegramStructureError,
  error_map,
  error_register_map,
)
from pylabrobot.storage.cytomat.schemas import (
  ActionRegisterState,
  OverviewRegisterState,
  SensorStates,
  SwapStationState,
)
from pylabrobot.storage.cytomat.utils import (
  hex_to_base_twelve,
  hex_to_binary,
  validate_storage_location_number,
)

logger = logging.getLogger(__name__)


class CytomatBackend(IncubatorBackend):
  default_baud = 9600
  serial_message_encoding = "utf-8"

  def __init__(self, model: Union[CytomatType, str], port: str):
    if not HAS_SERIAL:
      raise RuntimeError(
        "pyserial is not installed. Install with: pip install pylabrobot[serial]. "
        f"Import error: {_SERIAL_IMPORT_ERROR}"
      )
    super().__init__()

    supported_models = [
      CytomatType.C6000,
      CytomatType.C6002,
      CytomatType.C2C_425,
      CytomatType.C2C_450_SHAKE,
      CytomatType.C5C,
    ]
    if isinstance(model, str):
      try:
        model = CytomatType(model)
      except ValueError:
        raise ValueError(f"Unsupported Cytomat model: '{model}'")
    if model not in supported_models:
      raise NotImplementedError(
        f"Only the following Cytomats are supported: {supported_models}, but got '{model}'"
      )
    self.model = model
    self._racks: List[PlateCarrier] = []

    self.io = Serial(
      port=port,
      baudrate=self.default_baud,
      bytesize=serial.EIGHTBITS,
      parity=serial.PARITY_NONE,
      stopbits=serial.STOPBITS_ONE,
      write_timeout=1,
      timeout=1,
      human_readable_device_name="Cytomat",
    )

  async def setup(self):
    await self.io.setup()
    await self.initialize()
    await self.wait_for_task_completion()

  async def set_racks(self, racks: List[PlateCarrier]):
    await super().set_racks(racks)
    warnings.warn("Cytomat racks need to be configured with the exe software")

  async def stop(self):
    await self.io.stop()

  def _assemble_command(self, command_type: str, command: str, params: str):
    carriage_return = "\r" if self.model == CytomatType.C2C_425 else "\r\n"
    command = f"{command_type}:{command} {params}".strip() + carriage_return
    return command

  async def send_command(self, command_type: str, command: str, params: str) -> str:
    async def _send_command(command_str) -> str:
      logger.debug(command_str.encode(self.serial_message_encoding))
      await self.io.write(command_str.encode(self.serial_message_encoding))
      resp = (await self.io.read(128)).decode(self.serial_message_encoding)
      if len(resp) == 0:
        raise RuntimeError("Cytomat did not respond to command, is it turned on?")
      key, *values = resp.split()
      value = " ".join(values)

      if key == CytomatActionResponse.OK.value or key == command:
        # actions return an OK response, while checks return the command at the start of the response
        return value
      if key == CytomatActionResponse.ERROR.value:
        logger.error("Command %s failed with: '%s'", command_str, resp)
        if value == "03":
          error_register = await self.get_error_register()
          await self.reset_error_register()
          raise CytomatTelegramStructureError(f"Telegram structure error: {error_register}")
        if int(value, base=16) in error_map:
          await self.reset_error_register()
          raise error_map[int(value, base=16)]
        await self.reset_error_register()
        raise Exception(f"Unknown cytomat error code in response: {resp}")

      logger.error("Command %s received an unknown response: '%s'", command_str, resp)
      await self.reset_error_register()
      raise Exception(f"Unknown response from cytomat: {resp}")

    # Cytomats sometimes return a busy or command not recognized error even when the overview
    # register says the machine is not busy, or if the command is known. We will retry a few times,
    # which costs 1s if there is a true error, but is necessary to avoid false negatives.
    command_str = self._assemble_command(command_type=command_type, command=command, params=params)
    n_retries = 10
    exc: Optional[BaseException] = None
    for _ in range(n_retries):
      try:
        return await _send_command(command_str)
      except (CytomatCommandUnknownError, CytomatBusyError) as e:
        exc = e
        await asyncio.sleep(0.1)
        continue
    assert exc is not None
    await self.reset_error_register()
    raise exc

  async def send_action(
    self, command_type: str, command: str, params: str, timeout: Optional[int] = 60
  ) -> OverviewRegisterState:
    """Calls send_command, but has a timeout handler and returns the overview register state.
    Args:
      timeout: The maximum time to wait for the command to complete. If None, the command will not
        wait for completion.
    """
    await self.send_command(command_type, command, params)
    if timeout is not None:
      overview_register = await self.wait_for_task_completion(timeout=timeout)
    return overview_register

  def _site_to_firmware_string(self, site: PlateHolder) -> str:
    rack = cast(PlateCarrier, site.parent)
    rack_idx = [rack.name for rack in self._racks].index(
      rack.name
    )  # autoreload resistant, should work
    site_idx = next(idx for idx, s in rack.sites.items() if s == site)

    if self.model in [CytomatType.C2C_425]:
      return f"{str(rack_idx).zfill(2)} {str(site_idx).zfill(2)}"

    # TODO: configure all cytomats to use `rack site` format
    if self.model in [
      CytomatType.C6000,
      CytomatType.C6002,
      CytomatType.C2C_450_SHAKE,
      CytomatType.C5C,
    ]:
      slots_to_skip = sum(r.capacity for r in self._racks[:rack_idx])
      absolute_slot = slots_to_skip + site_idx + 1  # 1-indexed

      return f"{absolute_slot:03}"

    raise ValueError(f"Unsupported Cytomat model: {self.model}")

  async def get_overview_register(self) -> OverviewRegisterState:
    # Sometimes this command is not recognized and it is not known why. We will retry a few times
    # We don't care if the cytomat is still busy, that is actually what we are often checking for.
    # We are just gathering state, so just try a little bit later.
    num_tries = 10
    for _ in range(num_tries):
      try:
        resp = await self.send_command("ch", "bs", "")
      except (CytomatCommandUnknownError, CytomatBusyError):
        await asyncio.sleep(0.1)
        continue
      return OverviewRegisterState.from_resp(resp)
    await self.reset_error_register()
    raise CytomatCommandUnknownError("Could not get overview register")

  async def get_warning_register(self) -> WarningRegister:
    hex_value = await self.send_command("ch", "bw", "")
    for member in WarningRegister:
      if hex_value == member.value:
        return member

    await self.reset_error_register()
    raise Exception(f"Unknown warning register value: {hex_value}")

  async def get_error_register(self) -> ErrorRegister:
    hex_value = await self.send_command("ch", "be", "")
    for member in ErrorRegister:
      if hex_value == member.value:
        return member

    await self.reset_error_register()
    raise Exception(f"Unknown error register value: {hex_value}")

  async def reset_error_register(self) -> None:
    await self.send_command("rs", "be", "")

  async def initialize(self) -> None:
    await self.send_action("ll", "in", "", timeout=300)  # this command sometimes times out

  async def open_door(self):
    return await self.send_action("ll", "gp", "002")

  async def close_door(self):
    return await self.send_action("ll", "gp", "001")

  async def shovel_in(self):
    return await self.send_action("ll", "sp", "001")

  async def shovel_out(self):
    return await self.send_action("ll", "sp", "002")

  async def get_action_register(self) -> ActionRegisterState:
    hex_value = await self.send_command("ch", "ba", "")
    binary_repr = hex_to_binary(hex_value)
    target, action = binary_repr[:3], binary_repr[3:]

    target_enum = None
    for action_type_member in ActionType:
      if int(target, 2) == int(action_type_member.value, 16):
        target_enum = action_type_member
        break
    assert target_enum is not None, f"Unknown target value: {target}"

    action_enum = None
    for action_register_member in ActionRegister:
      if int(action, base=2) == int(action_register_member.value, base=16):
        action_enum = action_register_member
        break
    assert action_enum is not None, f"Unknown value: {action}"

    return ActionRegisterState(target=target_enum, action=action_enum)

  async def get_swap_register(self) -> SwapStationState:
    value = await self.send_command("ch", "sw", "")
    return SwapStationState(
      position=SwapStationPosition(int(value[0])),
      load_status_front_of_gate=LoadStatusFrontOfGate(int(value[1])),
      load_status_at_processor=LoadStatusAtProcessor(int(value[2])),
    )

  async def get_sensor_register(self) -> SensorStates:
    hex_value = await self.send_command("ch", "ts", "")
    binary_values = hex_to_base_twelve(hex_value)
    return SensorStates(
      **{member.name: bool(int(binary_values[member.value])) for member in SensorRegister}
    )

  async def action_transfer_to_storage(  # used by insert_plate
    self, site: PlateHolder
  ) -> OverviewRegisterState:
    """Open lift door, retrieve from transfer, close door, place at storage"""
    return await self.send_action("mv", "ts", self._site_to_firmware_string(site), timeout=120)

  async def action_storage_to_transfer(  # used by retrieve_plate
    self, site: PlateHolder
  ) -> OverviewRegisterState:
    """Retrieve from storage, open door, move to transfer, close door"""
    return await self.send_action("mv", "st", self._site_to_firmware_string(site))

  async def action_storage_to_wait(self, site: PlateHolder) -> OverviewRegisterState:
    """Retrieve from storage, move to wait position"""
    return await self.send_action("mv", "sw", self._site_to_firmware_string(site))

  async def action_wait_to_storage(self, site: PlateHolder) -> OverviewRegisterState:
    """Move from wait to storage, unload, return to wait"""
    return await self.send_action("mv", "ws", self._site_to_firmware_string(site))

  async def action_wait_to_transfer(self) -> OverviewRegisterState:
    """Open door, place on transfer, return to wait, close door"""
    return await self.send_action("mv", "wt", "")

  async def action_transfer_to_wait(self) -> OverviewRegisterState:
    """Open door, retrieve from transfer, return to wait, close door"""
    return await self.send_action("mv", "tw", "")

  async def action_wait_to_exposed(self) -> OverviewRegisterState:
    """Move from wait to exposed position outside device"""
    return await self.send_action("mv", "wh", "")

  async def action_exposed_to_wait(self) -> OverviewRegisterState:
    """Return to wait from exposed, close door"""
    return await self.send_action("mv", "hw", "")

  async def action_exposed_to_storage(self, site: PlateHolder) -> OverviewRegisterState:
    """Return with MTP from exposed to storage, move to wait, close door"""
    return await self.send_action("mv", "hs", self._site_to_firmware_string(site))

  async def action_storage_to_exposed(self, site: PlateHolder) -> OverviewRegisterState:
    """Move from wait to storage, load MTP, transport to exposed"""
    return await self.send_action("mv", "sh", self._site_to_firmware_string(site))

  async def action_read_barcode(
    self,
    site_number_a: str,
    site_number_b: str,
  ) -> OverviewRegisterState:
    # Read barcode of storage locations
    validate_storage_location_number(site_number_a)
    validate_storage_location_number(site_number_b)
    resp = await self.send_command("mv", "sn", f"{site_number_a} {site_number_b}")
    return OverviewRegisterState.from_resp(resp)

  async def wait_for_transfer_station(self, occupied: bool = False):
    """Wait for the transfer station to be occupied, or unoccupied."""
    while (await self.get_overview_register()).transfer_station_occupied != occupied:
      await asyncio.sleep(1)

  async def wait_for_task_completion(self, timeout=60) -> OverviewRegisterState:
    """
    Wait for the cytomat to finish the current task. This is done by checking the overview register
    until the busy bit is not set. If the cytomat is busy for too long, a TimeoutError is raised.
    If the error bit is set in the overview register, the error register is read and the corresponding
    error is raised.
    """
    start = time.time()
    while True:
      overview_register = await self.get_overview_register()
      if not overview_register.busy_bit_set:
        # only check for errors once the cytomat is done, so that the user has the chance to
        # handle the error and proceed if desired.
        if overview_register.error_register_set:
          error_register = await self.get_error_register()
          await self.reset_error_register()
          raise error_register_map[error_register]
        return overview_register
      await asyncio.sleep(1)
      if time.time() - start > timeout:
        raise TimeoutError("Cytomat did not complete task in time")

  async def init_shakers(self):
    return hex_to_binary(await self.send_command("ll", "vi", ""))

  async def start_shaking(self, frequency: float, shakers: Optional[List[int]] = None):
    if self.model == CytomatType.C5C:
      raise NotImplementedError("Shaking is not supported on this model")
    await self.set_shaking_frequency(frequency=int(frequency), shakers=shakers)
    return hex_to_binary(await self.send_command("ll", "va", ""))

  async def stop_shaking(self):
    if self.model == CytomatType.C5C:
      raise NotImplementedError("Shaking is not supported on this model")
    return hex_to_binary(await self.send_command("ll", "vd", ""))

  async def set_shaking_frequency(
    self, frequency: int, shakers: Optional[List[int]] = None
  ) -> List[str]:
    shakers = shakers or [1, 2]
    assert all(shaker in [1, 2] for shaker in shakers), "Shaker index must be 1 or 2"
    return [await self.send_command("se", f"pb 2{idx - 1}", f"{frequency:04}") for idx in shakers]

  async def get_incubation_query(
    self, query: Literal["ic", "ih", "io", "it"]
  ) -> CytomatIncupationResponse:
    resp = await self.send_command("ch", query, "")
    nominal, actual = resp.split()
    return CytomatIncupationResponse(
      nominal_value=float(nominal.lstrip("+")), actual_value=float(actual.lstrip("+"))
    )

  async def get_co2(self) -> CytomatIncupationResponse:
    return await self.get_incubation_query("ic")

  async def get_humidity(self) -> CytomatIncupationResponse:
    return await self.get_incubation_query("ih")

  async def get_o2(self) -> CytomatIncupationResponse:
    return await self.get_incubation_query("io")

  async def get_temperature(self) -> float:
    return (await self.get_incubation_query("it")).actual_value

  async def fetch_plate_to_loading_tray(self, plate: Plate, **backend_kwargs):
    site = plate.parent
    assert isinstance(site, PlateHolder)
    await self.action_storage_to_transfer(site)

  async def take_in_plate(self, plate: Plate, site: PlateHolder, **backend_kwargs):
    await self.action_transfer_to_storage(site)

  async def set_temperature(self, *args, **kwargs):
    raise NotImplementedError("Temperature control is not implemented yet")

  def serialize(self) -> dict:
    return {
      **IncubatorBackend.serialize(self),
      "model": self.model.value,
      "port": self.io.port,
    }


class CytomatChatterbox(CytomatBackend):
  async def setup(self):
    await self.wait_for_task_completion()

  async def stop(self):
    print("closing connection to cytomat")

  async def send_command(self, command_type, command, params):
    print(
      "cytomat", self._assemble_command(command_type=command_type, command=command, params=params)
    )
    if command_type == "ch":
      return "0"
    return "0" * 8

  async def wait_for_transfer_station(self, occupied: bool = False):
    # send the command, but don't wait when we are in chatting mode.
    _ = await self.get_overview_register()


# Deprecated alias with warning # TODO: remove mid May 2025 (giving people 1 month to update)
# https://github.com/PyLabRobot/pylabrobot/issues/466


class Cytomat:
  def __init__(self, *args, **kwargs):
    raise RuntimeError("`Cytomat` is deprecated. Please use `CytomatBackend` instead. ")



"""Definition file for various version numbers."""

import os

# Version number for pylabrobot
_version_file = os.path.join(os.path.dirname(__file__), "version.txt")
with open(_version_file, "r", encoding="utf-8") as f:
  __version__ = f.read().strip()

# Version number of the standard form protocol used by the server.
STANDARD_FORM_JSON_VERSION = "0.1.0"



import asyncio
from typing import Optional, Union

from pylabrobot.machines.machine import Machine

from .backend import PumpBackend
from .calibration import PumpCalibration


class Pump(Machine):
  """Frontend for a (peristaltic) pump."""

  def __init__(
    self,
    backend: PumpBackend,
    calibration: Optional[PumpCalibration] = None,
  ):
    super().__init__(backend=backend)
    self.backend: PumpBackend = backend  # fix type
    if calibration is not None and len(calibration) != 1:
      raise ValueError("Calibration may only have a single item for this pump")
    self.calibration = calibration

  def serialize(self) -> dict:
    if self.calibration is None:
      return super().serialize()
    return {
      **super().serialize(),
      "calibration": self.calibration.serialize(),
    }

  @classmethod
  def deserialize(cls, data: dict):
    data_copy = data.copy()
    calibration_data = data_copy.pop("calibration", None)
    if calibration_data is not None:
      calibration = PumpCalibration.deserialize(calibration_data)
      data_copy["calibration"] = calibration
    return super().deserialize(data_copy)

  async def run_revolutions(self, num_revolutions: float):
    """Run a given number of revolutions. This method will return after the command has been sent,
    and the pump will run until `halt` is called.

    Args:
      num_revolutions: number of revolutions to run
    """

    self.backend.run_revolutions(num_revolutions=num_revolutions)

  async def run_continuously(self, speed: float):
    """Run continuously at a given speed. This method will return after the command has been sent,
    and the pump will run until `halt` is called.

    If speed is 0, the pump will be halted.

    Args:
      speed: speed in rpm/pump-specific units.
    """

    self.backend.run_continuously(speed=speed)

  async def run_for_duration(self, speed: Union[float, int], duration: Union[float, int]):
    """Run the pump at specified speed for the specified duration.

    Args:
      speed: speed in rpm/pump-specific units.
      duration: duration to run pump.
    """

    if duration < 0:
      raise ValueError("Duration must be positive.")
    await self.run_continuously(speed=speed)
    await asyncio.sleep(duration)
    await self.run_continuously(speed=0)

  async def pump_volume(self, speed: Union[float, int], volume: Union[float, int]):
    """Run the pump at specified speed for the specified volume. Note that this function requires
    the pump to be calibrated at the input speed.

    Args:
      speed: speed in rpm/pump-specific units.
      volume: volume to pump.
    """

    if self.calibration is None:
      raise TypeError(
        "Pump is not calibrated. Volume based pumping and related functions unavailable."
      )
    if self.calibration.calibration_mode == "duration":
      duration = volume / self.calibration[0]
      await self.run_for_duration(speed=speed, duration=duration)
    elif self.calibration.calibration_mode == "revolutions":
      num_revolutions = volume / self.calibration[0]
      await self.run_revolutions(num_revolutions=num_revolutions)
    else:
      raise ValueError("Calibration mode not recognized.")

  async def halt(self):
    """Halt the pump."""
    self.backend.halt()



from typing import Optional

from pylabrobot.resources.height_volume_functions import (
  compute_height_from_volume_rectangle,
  compute_volume_from_height_rectangle,
)
from pylabrobot.resources.plate import Lid, Plate
from pylabrobot.resources.utils import create_ordered_items_2d
from pylabrobot.resources.well import (
  CrossSectionType,
  Well,
  WellBottomType,
)


def AGenBio_96_wellplate_Ub_2200ul(name: str, lid: Optional[Lid] = None) -> Plate:
  """
  AGenBio Catalog No. P-2.2-SQG-96
  - Material: Polypropylene
  - Max. volume: 2200 uL
  """
  INNER_WELL_WIDTH = 8  # measured
  INNER_WELL_LENGTH = 8  # measured

  well_kwargs = {
    "size_x": INNER_WELL_WIDTH,  # measured
    "size_y": INNER_WELL_LENGTH,  # measured
    "size_z": 38.2,  # measured to bottom of well
    "bottom_type": WellBottomType.FLAT,
    "cross_section_type": CrossSectionType.RECTANGLE,
    "compute_height_from_volume": lambda liquid_volume: compute_height_from_volume_rectangle(
      liquid_volume,
      INNER_WELL_LENGTH,
      INNER_WELL_WIDTH,
    ),
    "compute_volume_from_height": lambda liquid_height: compute_volume_from_height_rectangle(
      liquid_height,
      INNER_WELL_LENGTH,
      INNER_WELL_WIDTH,
    ),
    "material_z_thickness": 1,
  }

  return Plate(
    name=name,
    size_x=127.76,  # from spec
    size_y=85.48,  # from spec
    size_z=42.5,  # from spec
    lid=lid,
    model=AGenBio_96_wellplate_Ub_2200ul.__name__,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=9.38,  # measured
      dy=6.24,  # measured
      dz=3.8,  # measured
      item_dx=9,
      item_dy=9,
      **well_kwargs,
    ),
  )


def AGenBio_4_troughplate_75000uL_Vb(name: str, lid: Optional[Lid] = None) -> Plate:
  """
  AGenBio Catalog No. RES-75-4MW
  - Material: Polypropylene
  - Max. volume: 75 mL
  """
  INNER_WELL_WIDTH = 26.1  # measured
  INNER_WELL_LENGTH = 71.2  # measured

  well_kwargs = {
    "size_x": 26,  # measured
    "size_y": 71.2,  # measured
    "size_z": 42.55,  # measured to bottom of well
    "bottom_type": WellBottomType.FLAT,
    "cross_section_type": CrossSectionType.RECTANGLE,
    "compute_height_from_volume": lambda liquid_volume: compute_height_from_volume_rectangle(
      liquid_volume,
      INNER_WELL_LENGTH,
      INNER_WELL_WIDTH,
    ),
    "compute_volume_from_height": lambda liquid_height: compute_volume_from_height_rectangle(
      liquid_height,
      INNER_WELL_LENGTH,
      INNER_WELL_WIDTH,
    ),
    "material_z_thickness": 1,
  }

  return Plate(
    name=name,
    size_x=127.76,  # from spec
    size_y=85.48,  # from spec
    size_z=43.80,  # measured
    lid=lid,
    model=AGenBio_4_troughplate_75000uL_Vb.__name__,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=4,
      num_items_y=1,
      dx=9.8,  # measured
      dy=7.2,  # measured
      dz=0.9,  # measured
      item_dx=INNER_WELL_WIDTH + 1,  # 1 mm wall thickness
      item_dy=INNER_WELL_LENGTH,
      **well_kwargs,
    ),
  )


def AGenBio_4_troughplate_75000_Vb(name: str, lid: Optional[Lid] = None) -> Plate:
  """Deprecated. Use :func:`AGenBio_4_troughplate_75000uL_Vb` instead."""
  import warnings

  warnings.warn(
    "AGenBio_4_troughplate_75000_Vb is deprecated. Use AGenBio_4_troughplate_75000uL_Vb instead.",
    DeprecationWarning,
    stacklevel=2,
  )
  return AGenBio_4_troughplate_75000uL_Vb(name=name, lid=lid)


def AGenBio_1_wellplate_Fl(name: str, lid: Optional[Lid] = None) -> Plate:
  """Deprecated. Use :func:`AGenBio_1_troughplate_190000uL_Fl` instead."""
  import warnings

  warnings.warn(
    "AGenBio_1_wellplate_Fl is deprecated. Use AGenBio_1_troughplate_190000uL_Fl instead.",
    DeprecationWarning,
    stacklevel=2,
  )
  return AGenBio_1_troughplate_190000uL_Fl(name=name, lid=lid)


def AGenBio_1_troughplate_190000uL_Fl(name: str, lid: Optional[Lid] = None) -> Plate:
  """
  AGenBio Catalog No. RES-190-F
  - Material: Polypropylene
  - Max. volume: 190 mL
  """
  INNER_WELL_WIDTH = 107.2  # measured
  INNER_WELL_HEIGHT = 70.9  # measured

  well_kwargs = {
    "size_x": INNER_WELL_WIDTH,  # measured
    "size_y": INNER_WELL_HEIGHT,  # measured
    "size_z": 24.76,  # measured to bottom of well
    "bottom_type": WellBottomType.FLAT,
    "cross_section_type": CrossSectionType.RECTANGLE,
    "compute_height_from_volume": lambda liquid_volume: compute_height_from_volume_rectangle(
      liquid_volume,
      INNER_WELL_HEIGHT,
      INNER_WELL_WIDTH,
    ),
    "compute_volume_from_height": lambda liquid_height: compute_volume_from_height_rectangle(
      liquid_height,
      INNER_WELL_HEIGHT,
      INNER_WELL_WIDTH,
    ),
    "material_z_thickness": 1,
  }

  return Plate(
    name=name,
    size_x=127.76,  # from spec
    size_y=85.48,  # from spec
    size_z=44.2,  # measured
    lid=lid,
    model=AGenBio_1_troughplate_190000uL_Fl.__name__,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=1,
      num_items_y=1,
      dx=10.1,
      dy=7.6,
      dz=5.88,
      item_dx=INNER_WELL_WIDTH,
      item_dy=INNER_WELL_HEIGHT,
      **well_kwargs,
    ),
  )


def AGenBio_1_troughplate_100000uL_Fl(name: str, lid: Optional[Lid] = None) -> Plate:
  """
  AGenBio Catalog No. RES-100-F
  - Material: Polypropylene
  - Max. volume: 100 mL
  """
  INNER_WELL_WIDTH = 107.2  # measured
  INNER_WELL_HEIGHT = 70.9  # measured

  well_kwargs = {
    "size_x": INNER_WELL_WIDTH,  # measured
    "size_y": INNER_WELL_HEIGHT,  # measured
    "size_z": 13,  # measured to bottom of well
    "bottom_type": WellBottomType.FLAT,
    "cross_section_type": CrossSectionType.RECTANGLE,
    "compute_height_from_volume": lambda liquid_volume: compute_height_from_volume_rectangle(
      liquid_volume,
      INNER_WELL_HEIGHT,
      INNER_WELL_WIDTH,
    ),
    "material_z_thickness": 1,
  }

  return Plate(
    name=name,
    size_x=127.76,  # from spec
    size_y=85.48,  # from spec
    size_z=31.4,  # from spec
    lid=lid,
    model=AGenBio_1_troughplate_100000uL_Fl.__name__,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=1,
      num_items_y=1,
      dx=9.8,
      dy=7.6,
      dz=5.88,
      item_dx=INNER_WELL_WIDTH,
      item_dy=INNER_WELL_HEIGHT,
      **well_kwargs,
    ),
  )



from .access2 import Access2
from .centrifuge import Centrifuge, Loader
from .standard import (
  BucketHasPlateError,
  BucketNoPlateError,
  CentrifugeDoorError,
  LoaderNoPlateError,
  NotAtBucketError,
)
from .vspin_backend import Access2Backend, VSpinBackend



from pylabrobot.machines import Machine

from .backend import PeelerBackend


class Peeler(Machine):
  """A microplate peeler"""

  def __init__(self, backend: PeelerBackend):
    super().__init__(backend=backend)
    self.backend: PeelerBackend = backend

  async def peel(self, **backend_kwargs):
    return await self.backend.peel(**backend_kwargs)

  async def restart(self, **backend_kwargs):
    return await self.backend.restart(**backend_kwargs)



import unittest

from pylabrobot.resources import (
  CellVis_24_wellplate_3600uL_Fb,
  Cor_Cos_6_wellplate_16800ul_Fb,
  Revvity_384_wellplate_28ul_Ub,
  Thermo_TS_96_wellplate_1200ul_Rb,
)

from .coordinate import Coordinate
from .plate import Lid, Plate


class TestLid(unittest.TestCase):
  def test_initialize_with_lid(self):
    lid = Lid("plate_lid", size_x=1, size_y=1, size_z=10, nesting_z_height=10)
    plate = Plate(
      "plate",
      size_x=1,
      size_y=1,
      size_z=15,
      ordered_items={},
      lid=lid,
    )
    plate.location = Coordinate.zero()

    assert plate.lid is not None
    self.assertEqual(plate.lid.name, "plate_lid")
    self.assertEqual(plate.lid.get_absolute_size_x(), 1)
    self.assertEqual(plate.lid.get_absolute_location(), Coordinate(0, 0, 5))

  def _create_plate_with_lid(self):
    """Helper to create a plate with a lid."""
    plate = Plate("plate", size_x=1, size_y=1, size_z=1, ordered_items={})
    lid = Lid(
      name="another_lid",
      size_x=plate.get_size_x(),
      size_y=plate.get_size_y(),
      size_z=plate.get_size_z(),
      nesting_z_height=plate.get_size_z(),
    )
    plate.assign_child_resource(lid, location=Coordinate(0, 0, 0))
    return plate

  def test_add_lid(self):
    plate = self._create_plate_with_lid()
    self.assertIsNotNone(plate.lid)

  def test_add_lid_with_existing_lid(self):
    plate = self._create_plate_with_lid()
    another_lid = Lid(
      name="another_lid",
      size_x=plate.get_size_x(),
      size_y=plate.get_size_y(),
      size_z=plate.get_size_z(),
      nesting_z_height=plate.get_size_z(),
    )
    with self.assertRaises(ValueError):
      plate.assign_child_resource(another_lid, location=Coordinate(0, 0, 0))

    plate = self._create_plate_with_lid()
    plate.unassign_child_resource(plate.lid)
    self.assertIsNone(plate.lid)


class TestQuadrants(unittest.TestCase):
  def setUp(self):
    self.example_6_wellplate = Cor_Cos_6_wellplate_16800ul_Fb(name="example_6_wellplate")
    self.example_24_wellplate = CellVis_24_wellplate_3600uL_Fb(name="example_24_wellplate")
    self.example_96_wellplate = Thermo_TS_96_wellplate_1200ul_Rb(name="example_96_wellplate")
    self.example_384_wellplate = Revvity_384_wellplate_28ul_Ub(name="example_384_wellplate")

  def test_checkerboard_column_major(self):
    self.assertEqual(
      [
        well.get_identifier()
        for well in self.example_96_wellplate.get_quadrant(
          quadrant="tl", quadrant_type="checkerboard", quadrant_internal_fill_order="column-major"
        )
      ],
      [
        "A1",
        "C1",
        "E1",
        "G1",
        "A3",
        "C3",
        "E3",
        "G3",
        "A5",
        "C5",
        "E5",
        "G5",
        "A7",
        "C7",
        "E7",
        "G7",
        "A9",
        "C9",
        "E9",
        "G9",
        "A11",
        "C11",
        "E11",
        "G11",
      ],
    )

  def test_checkerboard_row_major(self):
    self.assertEqual(
      [
        well.get_identifier()
        for well in self.example_96_wellplate.get_quadrant(
          quadrant="tl", quadrant_type="checkerboard", quadrant_internal_fill_order="row-major"
        )
      ],
      [
        "A1",
        "A3",
        "A5",
        "A7",
        "A9",
        "A11",
        "C1",
        "C3",
        "C5",
        "C7",
        "C9",
        "C11",
        "E1",
        "E3",
        "E5",
        "E7",
        "E9",
        "E11",
        "G1",
        "G3",
        "G5",
        "G7",
        "G9",
        "G11",
      ],
    )

  def test_block_column_major(self):
    self.assertEqual(
      [
        well.get_identifier()
        for well in self.example_96_wellplate.get_quadrant(
          quadrant="tl", quadrant_type="block", quadrant_internal_fill_order="column-major"
        )
      ],
      [
        "A1",
        "B1",
        "C1",
        "D1",
        "A2",
        "B2",
        "C2",
        "D2",
        "A3",
        "B3",
        "C3",
        "D3",
        "A4",
        "B4",
        "C4",
        "D4",
        "A5",
        "B5",
        "C5",
        "D5",
        "A6",
        "B6",
        "C6",
        "D6",
      ],
    )

  def test_block_row_major(self):
    self.assertEqual(
      [
        well.get_identifier()
        for well in self.example_96_wellplate.get_quadrant(
          quadrant="tl", quadrant_type="block", quadrant_internal_fill_order="row-major"
        )
      ],
      [
        "A1",
        "A2",
        "A3",
        "A4",
        "A5",
        "A6",
        "B1",
        "B2",
        "B3",
        "B4",
        "B5",
        "B6",
        "C1",
        "C2",
        "C3",
        "C4",
        "C5",
        "C6",
        "D1",
        "D2",
        "D3",
        "D4",
        "D5",
        "D6",
      ],
    )



"""A hybrid between pylabrobot.shaking and pylabrobot.temperature_controlling"""

from pylabrobot.heating_shaking.backend import HeaterShakerBackend
from pylabrobot.heating_shaking.bioshake_backend import BioShake
from pylabrobot.heating_shaking.chatterbox import HeaterShakerChatterboxBackend
from pylabrobot.heating_shaking.hamilton_backend import (
  HamiltonHeaterShakerBackend,
  HamiltonHeaterShakerBox,
)
from pylabrobot.heating_shaking.heater_shaker import HeaterShaker
from pylabrobot.heating_shaking.inheco.thermoshake import (
  inheco_thermoshake,
  inheco_thermoshake_ac,
  inheco_thermoshake_rm,
)
from pylabrobot.heating_shaking.inheco.thermoshake_backend import InhecoThermoshakeBackend



import math
from typing import List, Optional

from pylabrobot.machines import Machine
from pylabrobot.resources import Coordinate, Plate
from pylabrobot.resources.resource_holder import ResourceHolder
from pylabrobot.resources.well import CrossSectionType, Well

from .tilter_backend import TilterBackend


class Tilter(ResourceHolder, Machine):
  """Resources that tilt plates."""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    backend: TilterBackend,
    hinge_coordinate: Coordinate,
    child_location: Coordinate,
    category: Optional[str] = None,
    model: Optional[str] = None,
  ):
    ResourceHolder.__init__(
      self,
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      category=category,
      model=model,
      child_location=child_location,
    )
    Machine.__init__(self, backend=backend)
    self.backend: TilterBackend = backend  # fix type
    self._absolute_angle: float = 0
    self._hinge_coordinate = hinge_coordinate

  @property
  def absolute_angle(self) -> float:
    return self._absolute_angle

  @property
  def hinge_coordinate(self) -> Coordinate:
    return self._hinge_coordinate

  async def set_angle(self, absolute_angle: float):
    """Set the tilt module to rotate to a given angle.

    Args:
      absolute_angle: The absolute (unsigned) angle to set rotation to, in degrees, measured from
        horizontal as zero.
    """

    await self.backend.set_angle(angle=absolute_angle)
    self._absolute_angle = absolute_angle

  def experimental_rotate_coordinate_around_hinge(
    self, absolute_coordinate: Coordinate, angle: float
  ) -> Coordinate:
    """Rotate an absolute coordinate around the hinge of the tilter by a given angle.

    Args:
      absolute_coordinate: The coordinate to rotate.
      angle: The angle to rotate by, in degrees. Negative is clockwise.

    Returns:
      Coordinate: The new coordinate after rotation.
    """
    theta = math.radians(angle)

    rotation_arm_x = absolute_coordinate.x - (
      self._hinge_coordinate.x + self.get_absolute_location("l", "f", "b").x
    )
    rotation_arm_z = absolute_coordinate.z - (
      self._hinge_coordinate.z + self.get_absolute_location("l", "f", "b").z
    )

    x_prime = rotation_arm_x * math.cos(theta) - rotation_arm_z * math.sin(theta)
    z_prime = rotation_arm_x * math.sin(theta) + rotation_arm_z * math.cos(theta)

    new_x = x_prime + (self._hinge_coordinate.x + self.get_absolute_location("l", "f", "b").x)
    new_z = z_prime + (self._hinge_coordinate.z + self.get_absolute_location("l", "f", "b").z)

    return Coordinate(new_x, absolute_coordinate.y, new_z)

  def experimental_get_plate_drain_offsets(
    self, plate: Plate, absolute_angle: Optional[float] = None
  ) -> List[Coordinate]:
    """Get the drain edge offsets for all wells in the given plate, tilted around the hinge at a
    given absolute angle.

    Args:
      plate: The plate to calculate the offsets for.
      absolute_angle: The absolute angle to rotate the plate. If `None`, the current tilt angle.
    """

    if absolute_angle is None:
      absolute_angle = self._absolute_angle
    assert absolute_angle is not None  # mypy
    angle = absolute_angle if self._hinge_coordinate.x < self._size_x / 2 else -absolute_angle

    _hinge_side = "l" if self._hinge_coordinate.x < self._size_x / 2 else "r"

    well_drain_offsets = []
    for well in plate.children:
      level_absolute_well_drain_coordinate = well.get_absolute_location(_hinge_side, "c", "b")
      rotated_absolute_well_drain_coordinate = self.experimental_rotate_coordinate_around_hinge(
        level_absolute_well_drain_coordinate, angle
      )
      well_drain_offset = rotated_absolute_well_drain_coordinate - well.get_absolute_location(
        "c", "c", "b"
      )
      well_drain_offsets.append(well_drain_offset)

    return well_drain_offsets

  def experimental_get_well_drain_offsets(
    self,
    wells: List[Well],
    n_tips: int = 1,
    absolute_angle: Optional[float] = None,
  ) -> List[Coordinate]:
    """Get the drain edge offsets for the given wells, tilted around the hinge at a
    given absolute angle, for multiple tips.

    Args:
      wells: The wells to calculate the offsets for.
      n_tips: The number of tips to calculate offsets for. Defaults to 1.
      absolute_angle: The absolute angle to rotate the wells. If `None`, the current tilt angle.

    Returns:
      A list of lists of Coordinates, where each inner list contains the offsets for n_tips.
    """

    if absolute_angle is None:
      absolute_angle = self._absolute_angle
    assert absolute_angle is not None  # mypy
    angle = absolute_angle * (-1 if self._hinge_coordinate.x >= self._size_x / 2 else 1)

    hinge_on_left = self._hinge_coordinate.x < self._size_x / 2
    min_tip_distance = 9  # mm

    well_drain_offsets = []
    for well in wells:
      assert well.cross_section_type == CrossSectionType.CIRCLE, (
        "Wells must have circular cross-section"
      )

      diameter = well.get_absolute_size_x()  # assuming circular well
      radius = diameter / 2

      if n_tips > 1:
        assert (n_tips - 1) * min_tip_distance <= diameter, (
          f"Cannot fit {n_tips} tips in a well with diameter {diameter} mm"
        )

        y_offsets = [
          ((n_tips - 1) / 2 - tip_index) * min_tip_distance for tip_index in range(n_tips)
        ]

        x_offset = math.sqrt(radius**2 - max(y_offsets) ** 2)
        x_offset = -x_offset if hinge_on_left else x_offset

        tip_coords = [Coordinate(x_offset, y, 0) for y in y_offsets]
      else:
        # Default case: n_tips = 1
        x_offset = -radius if hinge_on_left else radius
        tip_coords = [Coordinate(x_offset, 0, 0)]

      offsets = []
      for tip_coord in tip_coords:
        rotated_tip = self.experimental_rotate_coordinate_around_hinge(
          well.get_absolute_location("c", "c", "b") + tip_coord,
          angle,
        )
        offset = rotated_tip - well.get_absolute_location("c", "c", "b")
        offsets.append(offset)

      well_drain_offsets.append(offsets)

    return [offset for well_offsets in well_drain_offsets for offset in well_offsets]

  async def tilt(self, relative_angle: float):
    """Tilt the plate contained in the tilt module by a given angle relative to the current angle.

    Args:
      relative_angle: The angle to rotate by, in degrees. Clockwise. 0 is horizontal.
    """
    await self.set_angle(self._absolute_angle + relative_angle)



from pylabrobot.liquid_handling.backends.hamilton.STAR_backend import STAR
from pylabrobot.resources import Coordinate, Resource


class Pump(Resource):
  """Pump is the washer. Will rename later."""

  def __init__(self, backend: STAR, name="pump"):
    super().__init__(name, size_x=1, size_y=1, size_z=1, category="pump")
    self.backend = backend

    # assign two chambers to the pump
    self.chamber_1 = Resource("chamber_1", size_x=121, size_y=85, size_z=2, category="chamber")
    self.assign_child_resource(self.chamber_1, location=Coordinate(18.05, 371.500 - 63, 99))
    self.chamber_2 = Resource("chamber_2", size_x=121, size_y=85, size_z=2, category="chamber")
    self.assign_child_resource(self.chamber_2, location=Coordinate(18.05, 241.500 - 63, 100))

  async def refill(self):
    await self.backend.drain_dual_chamber_system(pump_station=1)
    await self.backend.fill_selected_dual_chamber(
      pump_station=1,
      drain_before_refill=False,
      wash_fluid=2,
      chamber=1,
      waste_chamber_suck_time_after_sensor_change=0,
    )
    await self.backend.fill_selected_dual_chamber(
      pump_station=1,
      drain_before_refill=False,
      wash_fluid=1,
      chamber=2,
      waste_chamber_suck_time_after_sensor_change=0,
    )



import unittest

from .linalg import matrix_multiply_3x3, matrix_vector_multiply_3x3


class TestLinalg(unittest.TestCase):
  """Test the linalg functions"""

  def test_matrix_multiply_3x3(self):
    A = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    B = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    C = matrix_multiply_3x3(A, B)
    assert C == [[30, 36, 42], [66, 81, 96], [102, 126, 150]]

  def test_matrix_vector_multiply_3x3(self):
    A = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    B = [1, 2, 3]
    C = matrix_vector_multiply_3x3(A, B)
    assert C == [14, 32, 50]



from pathlib import Path
from typing import Union

from pylabrobot.config.config import Config
from pylabrobot.config.io import ConfigReader, ConfigWriter


class FileReader(ConfigReader):
  """A ConfigReader that reads from a file."""

  encoding = "utf-8"

  def read(self, r: Union[str, Path]) -> Config:
    """Read a Config object from a file."""
    with open(r, self.open_mode, encoding=self.encoding) as f:
      return self.format_loader.load(f)


class FileWriter(ConfigWriter):
  """A ConfigWriter that writes to a file."""

  encoding = "utf-8"

  def write(self, w: Union[str, Path], cfg: Config):
    """Write a Config object to a file."""
    with open(w, self.open_mode, encoding=self.encoding) as f:
      self.format_saver.save(f, cfg)



import unittest

from pylabrobot.resources.errors import (
  TooLittleLiquidError,
  TooLittleVolumeError,
)
from pylabrobot.resources.volume_tracker import VolumeTracker


class TestVolumeTracker(unittest.TestCase):
  """Test for the tip volume tracker"""

  def test_init(self):
    tracker = VolumeTracker(thing="test", max_volume=100)
    self.assertEqual(tracker.get_free_volume(), 100)
    self.assertEqual(tracker.get_used_volume(), 0)

    tracker.set_volume(20)
    self.assertEqual(tracker.get_free_volume(), 80)
    self.assertEqual(tracker.get_used_volume(), 20)

  def test_add_liquid(self):
    tracker = VolumeTracker(thing="test", max_volume=100)

    tracker.add_liquid(volume=20)
    self.assertEqual(tracker.get_used_volume(), 20)
    self.assertEqual(tracker.get_free_volume(), 80)

    tracker.commit()
    self.assertEqual(tracker.get_used_volume(), 20)
    self.assertEqual(tracker.get_free_volume(), 80)

    with self.assertRaises(TooLittleVolumeError):
      tracker.add_liquid(volume=100)

  def test_remove_liquid(self):
    tracker = VolumeTracker(thing="test", max_volume=100, initial_volume=60)
    tracker.commit()

    self.assertEqual(tracker.get_used_volume(), 60)
    tracker.remove_liquid(volume=20)
    tracker.commit()
    self.assertEqual(tracker.get_used_volume(), 40)

    with self.assertRaises(TooLittleLiquidError):
      tracker.remove_liquid(volume=100)
